from math import *
import numpy as np
from colour import Color
import httplib2
import string as st
from scipy.linalg import null_space
from httplib2 import *
import sys
import quadprog
from .types import *
from typing import Callable
import urllib.request
import io
import hashlib
import os
if os.environ['CPP_SO_FOUND']=="1":
    import uaibot_cpp_bind as ub_cpp
    

class Utils:
    """A library that contains some utilities for UAIbot. All of the functions are static."""

    #######################################
    # Constants
    #######################################

    UAIBOT_NAME_TYPES = ['uaibot.', 'cylinder.', 'box.', 'ball.', 'convexpolytope.', 'robot.', 'simulation.', 'meshmaterial.', 'mtlmeshmaterial.',
                             'glbmeshmaterial.', 'texture.', 'pointlight.', 'frame.', 'model3d.', 'links.', 'pointcloud.', 'arrow.', 'rigidobject.',
                             '.group', '.htmldiv', 'CPP_GeometricPrimitives', 'CPP_DistStructRobotObj','CPP_AABB', '.Pedestrian', '.ObstacleColumn', '.ObstacleThinWall']

    IS_SIMPLE = ['uaibot.Ball', 'uaibot.Box', 'uaibot.Cylinder', 'uaibot.ConvexPolytope']
    
    IS_METRIC = ['uaibot.Ball', 'uaibot.Box', 'uaibot.Cylinder', 'uaibot.PointCloud', 'uaibot.ConvexPolytope']

    IS_GROUPABLE = ['uaibot.Ball', 'uaibot.Box', 'uaibot.Cylinder', 'uaibot.ConvexPolytope', 'uaibot.Frame',
                    'uaibot.RigidObject', 'uaibot.Group', 'uaibot.Robot', 'uaibot.PointLight']

    IS_OBJ_SIM = ['uaibot.Ball', 'uaibot.Box', 'uaibot.Cylinder', 'uaibot.ConvexPolytope', 'uaibot.Robot',
                  'uaibot.PointLight', 'uaibot.Frame', 'uaibot.PointCloud', 'uaibot.Arrow',
                  'uaibot.RigidObject', 'uaibot.Group', 'uaibot.HTMLDiv', 'uaibot.Pedestrian', 'uaibot.ObstacleColumn', 'uaibot.ObstacleThinWall']

    #######################################
    # Basic functions
    #######################################

    @staticmethod
    def cvt(input: Matrix) -> np.matrix:
        if isinstance(input, list) or isinstance(input, tuple):
            if all(isinstance(x, (int, float)) for x in input):
                return np.matrix(input, dtype=np.float64).reshape(-1, 1)
            elif all(isinstance(x, list) for x in input):
                return np.matrix(input, dtype=np.float64)
            else:
                raise ValueError("Entry is not a valid matrix.")
        
        elif isinstance(input, np.ndarray):
            if input.ndim == 1:
                return np.matrix(input.reshape(-1, 1), dtype=np.float64)
            else:
                if input.shape[1] > 1:
                    if input.shape[0]==1:
                        return np.matrix(input, dtype=np.float64).reshape(-1, 1)
                    else:
                        return np.matrix(input, dtype=np.float64)
                else:
                    return np.matrix(input, dtype=np.float64)                    
        else:
            raise TypeError("Entry is not a valid matrix.")                  

        

    @staticmethod
    def S(v: Vector) -> np.matrix:
        """
      Returns a 3x3 matrix that implements the cross product for a 3D vector  
      as a matricial product, that is, a matrix S(v) such that for any other 
      3D column  vector w, S(v)w = cross(v,w).
      
      Parameters
      ----------
      v : a 3D vector (3-element list/tuple, (3,1)/(1,3)/(3,)-shaped numpy matrix/numpy array)
          The vector for which the S matrix will be created.

      Returns
      -------
      S : 3x3 numpy matrix
          A matrix that implements the cross product with v.
      """
        vv = Utils.cvt(v)
        return np.matrix([[0, -vv[2,0], vv[1,0]],
                         [vv[2,0], 0, -vv[0,0]],
                         [-vv[1,0], vv[0,0], 0]])

    @staticmethod
    def rot(axis: Vector, angle: float) -> HTMatrix:
        """
      Homogeneous transformation matrix that represents the rotation of an
      angle in an axis.
      
      Parameters
      ----------
      axis : a 3D vector (3-element list/tuple, (3,1)/(1,3)/(3,)-shaped numpy matrix/numpy array)
          The axis of rotation.
      
      angle: float
          The angle of rotation, in radians.

      Returns
      -------
      htm : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """
        a = Utils.cvt(axis)
        a = a / np.linalg.norm(a)
        K = Utils.S(a)
        Q = np.identity(3) + sin(angle) * K + (1 - cos(angle)) * (K * K)
        return np.hstack([np.vstack([Q, np.matrix([0, 0, 0])]), np.matrix([[0], [0], [0], [1]])])

    @staticmethod
    def trn(vector: Vector) -> HTMatrix:
        """
      Homogeneous transformation matrix that represents the displacement
      of a vector
      
      Parameters
      ----------
      vector : a 3D vector (3-element list/tuple, (3,1)/(1,3)/(3,)-shaped numpy matrix/numpy array)
          The displacement vector.
      
      Returns
      -------
      htm : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """
        v = Utils.cvt(vector)
        return np.matrix([[1, 0, 0, v[0,0]],
                         [0, 1, 0, v[1,0]],
                         [0, 0, 1, v[2,0]],
                         [0, 0, 0, 1]])

    @staticmethod
    def rotx(angle: float) -> HTMatrix:
        """
      Homogeneous transformation matrix that represents the rotation of an
      angle in the 'x' axis.
      
      Parameters
      ----------
      angle: float
          The angle of rotation, in radians.

      Returns
      -------
      htm : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """
        return np.matrix([[1, 0, 0, 0],
                         [0, cos(angle), -sin(angle), 0],
                         [0, sin(angle), cos(angle), 0],
                         [0, 0, 0, 1]])

    @staticmethod
    def roty(angle: float) -> HTMatrix:
        """
      Homogeneous transformation matrix that represents the rotation of an
      angle in the 'y' axis.
      
      Parameters
      ----------
      angle: float
          The angle of rotation, in radians.

      Returns
      -------
      htm : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """
        return np.matrix([[cos(angle), 0, sin(angle), 0],
                         [0, 1, 0, 0],
                         [-sin(angle), 0, cos(angle), 0],
                         [0, 0, 0, 1]])

    @staticmethod
    def rotz(angle: float) -> HTMatrix:
        """
      Homogeneous transformation matrix that represents the rotation of an
      angle in the 'z' axis.
      
      Parameters
      ----------
      angle: float
          The angle of rotation, in radians.

      Returns
      -------
      htm : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """
        return np.matrix([[cos(angle), -sin(angle), 0, 0],
                         [sin(angle), cos(angle), 0, 0],
                         [0, 0, 1, 0],
                         [0, 0, 0, 1]])

    @staticmethod
    def htm_rand(trn_min: float =[0.,0.,0.], trn_max: float = [1.,1.,1.], rot: float =np.pi/2) -> HTMatrix:
        """
      Returns a random homogeneous transformation matrix.

      Parameters
      ----------
      trn: float
          Maximum parameter for random translation in x, y and z.
          (default: 1)

      rot: float
          Maximum parameter for random rotation in x, y and z.
          (default: 1)

      Returns
      -------
      htm : 4x4 numpy matrix
          A homogeneous transformation matrix.
      """

        x = np.random.uniform(trn_min[0],trn_max[0])
        y = np.random.uniform(trn_min[1],trn_max[1])
        z = np.random.uniform(trn_min[2],trn_max[2])
        ax = np.random.uniform(-rot / 2, rot / 2)
        ay = np.random.uniform(-rot / 2, rot / 2)
        az = np.random.uniform(-rot / 2, rot / 2)

        return Utils.trn([x,y,z]) * Utils.rotx(ax) * Utils.roty(ay) * Utils.rotz(az)

    @staticmethod
    def inv_htm(htm: HTMatrix) -> HTMatrix:
        """
      Given a homogeneous transformation matrix, compute its inverse.
      It is faster than using numpy.linalg.inv in the case of HTMs.
      
      Parameters
      ----------
      htm: 4X4 numpy matrix
          Homogeneous transformation matrix of the rotation.

      Returns
      -------
      inv_htm: 4X4 numpy array
          The inverse of the transformation matrix.       
      """
        Q = htm[0:3, 0:3]
        p = htm[0:3, 3]

        inv_htm = np.matrix(np.zeros((4, 4)))
        inv_htm[0:3, 0:3] = Q.T
        inv_htm[0:3, 3] = -Q.T * p
        inv_htm[3, 3] = 1

        return inv_htm

    @staticmethod
    def axis_angle(htm: Matrix) -> Tuple[np.matrix, float]:
        """
      Given an homogeneous transformation matrix representing a rotation, 
      return the rotation axis angle.
      
      Parameters
      ----------
      htm: 4X4 numpy matrix
          Homogeneous transformation matrix of the rotation.

      Returns
      -------
      axis : 3D numpy vector
          The rotation axis.

      angle : float
          The rotation angle, in radians.        
      """
      
        Q = htm[0:3, 0:3]
        trace = Q[0, 0] + Q[1, 1] + Q[2, 2]
        angle = acos((trace - 1) / 2)
        G = Q * Q - 2 * cos(angle) * Q + np.identity(3)
        ok = False
        while not ok:
            v = np.matrix(np.random.uniform(-100, 100, size=(3,1)))
            w = np.matrix(np.random.uniform(-100, 100, size=(3,1)))
            r = G * v
            nr = np.linalg.norm(r)
            prod = w.T * r
            if nr > 0.01:
                ortr = w -  (r * prod) / (nr * nr)
                axis = Utils.S(ortr) * (Q * ortr)
                naxis = np.linalg.norm(axis)
                ok = naxis > 0.01

        axis = axis / naxis
        return axis, angle

    @staticmethod
    def euler_angles(htm: Matrix) -> Tuple[float, float, float]:
        """
        Computer the Euler angles of a rotation matrix.
        Find alpha, beta and gamma such that.

        htm = Utils.rotz(alpha) * Utils.roty(beta) * Utils.rotx(gamma).

        Parameters
        ----------
        htm: 4X4 numpy array or nested list
            Homogeneous transformation matrix of the rotation.

        Returns
        -------
        alpha : float
            Rotation in z, in radians.
        beta : float
            Rotation in y, in radians.
        gamma : float
            Rotation in x, in radians.
        """

        Q = np.matrix(htm[0:3, 0:3])
        sy = sqrt(Q[0, 0] ** 2 + Q[1, 0] ** 2)

        if abs(sy) > 0.001:
            gamma = np.arctan2(Q[2, 1], Q[2, 2])
            beta = np.arctan2(-Q[2, 0], sy)
            alpha = np.arctan2(Q[1, 0], Q[0, 0])
        else:
            gamma = np.arctan2(-Q[1, 2], Q[1, 1])
            beta = np.arctan2(-Q[2, 0], sy)
            alpha = 0

        return alpha, beta, gamma


    @staticmethod
    def dp_inv(A: Matrix, eps: float = 0.001) -> np.matrix:
        """
      Compute the damped pseudoinverse of the matrix 'mat'.
      
      Parameters
      ----------
      A: a matrix ((n,m)-element list/tuple, (n,m)-shaped numpy matrix/numpy array)
          The matrix to compute the damped pseudoinverse.
      
      eps: positive float
          The damping factor.
          (default: 0.001).

      Returns
      -------
      pinvA: mxn numpy array
          The damped pseudoinverse of 'mat'.
      """
        A_int =  Utils.cvt(A)
        n = np.shape(A_int)[1]
        return np.linalg.inv(A_int.T * A_int + eps * np.identity(n)) * A_int.T

    @staticmethod
    def dp_inv_solve(A: Matrix, b: Vector, eps: float = 0.001, mode: str ='auto') -> np.matrix:
        """
      Solve the problem of minimizing ||A*x-b||^2 + eps*||x||^2
      It is the same as dp_inv(A,eps)*b, but faster
      
      Parameters
      ----------
      A: a matrix ((n,m)-element list/tuple, (n,m)-shaped numpy matrix/numpy array)
          A Matrix.
          
      b: a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
          b Vector.
                
      eps: positive float
          The damping factor.
          (default: 0.001).

      Returns
      -------
      x: mx1 numpy array
          The solution to the optimization problem.
      """
      
        if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
            raise Exception("c++ mode is set, but .so file was not loaded!")

        A_cvt = Utils.cvt(A)
        b_cvt = Utils.cvt(b)
        n, m = A_cvt.shape
        
        if mode == 'python' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
            
            M = np.block([
                [eps * np.eye(m), -A_cvt.T],
                [A_cvt, np.eye(n)]
            ])
            
            rhs = np.concatenate((np.zeros((m,1)), b_cvt))
            solution = np.linalg.solve(M, rhs)
            
            return Utils.cvt(solution[:m])
        else:
            return Utils.cvt(ub_cpp.dp_inv_solve(A_cvt,b_cvt,eps))
                 

    @staticmethod
    def interpolate(points: List[Vector], is_closed : bool =False) -> Callable[[List[float]], List[np.matrix]]:
        """
      Create a function handle that generates an one-time differentiable interpolated data from 'points'.

      The simplest case in when 'points' is a list with m elements. In this case, it will output a function f.
      When this function is evaluated at a scalar t, it will coincide with points[i] when t = i/m, that is,
      f(i/m) = points[i]. This function is once differentiable and periodic with period 1, so f(t+k)=f(t) for
      an integer k.

      The function can also use a n x m numpy array or lists as 'points'. In this case, f(t) is a n x 1 
      numpy vector.

      Finally, t can be a list of k elements instead of just a scalar. In this case, f(t) is a list of n x 1 numpy 
      matrices, in which the k-th element is the same as computing f(t[k]).


      Parameters
      ----------
      points: a list of nD vectors (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
          Points to be interpolated.
          
      is_closed: bool
          If the curve is closed or not (that is, f(0)=f(1)).
          (default: False)

      Returns
      -------
      f: function handle
          The function handle that implements the interpolation. 
      """

        if not Utils.is_a_list_vector(points):
            raise Exception("The parameter 'points' should be a list of vectors with the same dimension.")


        def aux_interpolate_single(arg_points):
            def int_aux_simp(arg_t, arg_n, arg_c):
                tn = arg_n * (arg_t % 1 if is_closed else min(max(arg_t, 0), 1))
                ti = min(floor(tn), arg_n - 1)
                tloc = tn - ti
                coef = arg_c[4 * ti:4 * ti + 4]
                return coef[0] + coef[1] * tloc + coef[2] * tloc**2 + coef[3] * tloc**3

            def int_aux(arg_t, arg_n, arg_c):
                return [int_aux_simp(tt, arg_n, arg_c) for tt in arg_t] if isinstance(arg_t, list) else int_aux_simp(arg_t, arg_n, arg_c)

            n = len(arg_points)
            num_segments = n if is_closed else n - 1
            A_rows = []
            b_vals = []

            # Interpolation constraints (value at start and end of each segment)
            for i in range(num_segments):
                row_start = np.zeros(4 * num_segments)
                row_end = np.zeros(4 * num_segments)
                row_start[4 * i:4 * i + 4] = [1, 0, 0, 0]
                row_end[4 * i:4 * i + 4] = [1, 1, 1, 1]
                A_rows.append(row_start)
                b_vals.append(arg_points[i])
                A_rows.append(row_end)
                b_vals.append(arg_points[(i + 1) % n if is_closed else i + 1])

            # Derivative continuity at junctions (including wraparound if closed)
            for i in range(num_segments):
                i_next = (i + 1) % num_segments
                if not is_closed and i == num_segments - 1:
                    break  # skip final derivative constraint for open curve
                row = np.zeros(4 * num_segments)
                row[4 * i + 1:4 * i + 4] = [1, 2, 3]
                row[4 * i_next + 1:4 * i_next + 4] = [-1, 0, 0]
                A_rows.append(row)
                b_vals.append(0)

            # Assemble matrices
            A = np.vstack(A_rows)
            b = np.array(b_vals).reshape(-1, 1)

            # Regularizer: minimize integral of second derivative squared
            H = np.zeros((4 * num_segments, 4 * num_segments))
            for i in range(num_segments):
                H_i = np.array([
                    [0, 0, 0, 0],
                    [0, 0, 0, 0],
                    [0, 0, 4, 6],
                    [0, 0, 6, 12]
                ])
                H[4 * i:4 * i + 4, 4 * i:4 * i + 4] = H_i

            # Solve constrained minimization via KKT system
            m = A.shape[0]
            KKT = np.block([
                [H, A.T],
                [A, np.zeros((m, m))]
            ])
            rhs = np.vstack([np.zeros((4 * num_segments, 1)), b])
            sol = np.linalg.solve(KKT, rhs)

            c = sol[:4 * num_segments].flatten().tolist()
            return lambda s: int_aux(s, num_segments, c)


        def aux_interpolate_multiple(arg_points, t):

            if not (Utils.is_a_vector(t) or Utils.is_a_number(t)):
                raise Exception(
                    "The parameter of the interpolation function must be either a number or a list of numbers.")

            y = np.zeros((0, len(t) if Utils.is_a_vector(t) else 1))
            for i in range(np.shape(Utils.cvt(arg_points[0]))[0]):
                fun = aux_interpolate_single([Utils.cvt(p)[i,0] for p in arg_points])
                fun_out = fun(t)
                fun_out = np.array(fun_out).reshape((1, len(fun_out) if Utils.is_a_vector(fun_out) else 1))
                y = np.block([[y], [fun_out]])

            list_of_points = [np.matrix(y[:, i].reshape(-1, 1)) for i in range(y.shape[1])]
            if isinstance(t, list):
                return list_of_points
            else:
                return list_of_points[0]

        return lambda t: aux_interpolate_multiple(points, t)

    @staticmethod
    def solve_qp(H: Matrix, f: Vector, A: Optional[Matrix] = None, 
                 b: Optional[Vector] = None, A_eq: Optional[Matrix] = None, 
                 b_eq: Optional[Vector] = None) -> np.matrix:
        """
        Solve the convex quadratic optimization problem:
        min_u 0.5 * u'Hu + f'u such that A_eq * u = b_eq and A * u >= b.
        
        Parameters
        ----------
        H : (n,n)-shaped matrix (list/tuple/np.matrix/np.ndarray)
            The H matrix. Must be symmetric positive definite.

        f : (n,)-shaped vector (list/tuple/np.matrix/np.ndarray)
            The f vector.

        A : (m,n)-shaped matrix or 'None', optional
            Inequality constraint matrix (Au >= b). Pass 'None' to omit.

        b : (m,)-shaped vector or 'None', optional
            RHS of inequality constraints. Pass 'None' to omit.

        A_eq : (p,n)-shaped matrix or 'None', optional
            Equality constraint matrix (A_eq * u = b_eq). Pass 'None' to omit.

        b_eq : (p,)-shaped vector or 'None', optional
            RHS of equality constraints. Pass 'None' to omit.
        
        Returns
        -------
        u : (n,1) numpy matrix
            The solution.
        """
        
        if (A is None) and (A_eq is None):
            return -np.linalg.inv(Utils.cvt(H))*(Utils.cvt(f))
        
        H_cvt = np.matrix(H, dtype=np.float64)
        H_cvt = 0.5 * (H_cvt + H_cvt.T)  # Ensure symmetry
        f_cvt = np.array(np.asarray(f).reshape((-1,)), dtype=np.float64)

        # Handle empty A and b
        if (A is None) or (b is None):
            A_ineq = np.empty((0, H_cvt.shape[0]))
            b_ineq = np.empty((0,))
        else:
            A_ineq = np.matrix(A, dtype=np.float64)
            b_ineq = np.array(np.asarray(b).reshape((-1,)), dtype=np.float64)

        # Handle empty A_eq and b_eq
        if (A_eq is None) or (b_eq is None):
            A_eq_cvt = np.empty((0, H_cvt.shape[0]))
            b_eq_cvt = np.empty((0,))
        else:
            A_eq_cvt = np.matrix(A_eq, dtype=np.float64)
            b_eq_cvt = np.array(np.asarray(b_eq).reshape((-1,)), dtype=np.float64)

        # Stack constraints and compute meq
        G = np.vstack([A_eq_cvt, A_ineq])  
        h = np.hstack([b_eq_cvt, b_ineq])
        meq = A_eq_cvt.shape[0]

        result = quadprog.solve_qp(H_cvt, -f_cvt, G.T, h, meq)
        return Utils.cvt(result[0])

    @staticmethod
    def null_space(A: Matrix) -> np.matrix:
        """
        Compute the null space of the matrix A.

        Parameters:
        ----------
        A : a matrix ((n,m)-element list/tuple, (n,m)-shaped numpy matrix/numpy array)
            The matrix to compute the null space.

        Returns
        -------
        A_null : m x r numpy matrix
            The matrix such that its column spam the null space.
        """
        return np.matrix(null_space(Utils.cvt(A)))
    
    @staticmethod
    def get_fishbotics_mp_problems() -> dict:
        """
        Gets a set of many motion planning problems from the 'FishBotics' dataset.
        They were designed for the Franka Emika Panda robot.
        See: https://github.com/fishbotics/robometrics
        
        
        It is a dictionary in which each key is the name of the problem.
        If 'all_problems' is the output of this function, use
         list(all_problems.keys()) to get the keys (the problem names).
        
        Each problem is again a dictionary, which contains the following keys:
        
        'all_obs': a list of UAIBot objects representing the obstacles.
        'q0': the initial configuration.
        'htm_base': the initial pose for the base.
        'htm_tg': the target HTML for the end-effector.
        

        Parameters:
        ----------
        None.

        Returns
        -------
        all_problems : dictionary
            The dictionary with all the problems.
        """          
        
        
        def load_npz_from_url_cached(url, cache_dir=".cache"):
            os.makedirs(cache_dir, exist_ok=True)
            
            # Create a unique filename based on the URL hash
            url_hash = hashlib.sha256(url.encode()).hexdigest()
            local_path = os.path.join(cache_dir, f"{url_hash}.npz")

            # If cached, load from disk
            if os.path.exists(local_path):
                return np.load(local_path, allow_pickle=True)

            # Else, download and save
            with urllib.request.urlopen(url) as response:
                data = response.read()
            with open(local_path, 'wb') as f:
                f.write(data)
            
            return np.load(local_path, allow_pickle=True)


        allproblems_1 = load_npz_from_url_cached("https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/MotionPlanningProblems/fishbotics_mp_problems_part_1.npz")
        allproblems_2 = load_npz_from_url_cached("https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/MotionPlanningProblems/fishbotics_mp_problems_part_2.npz")
        allproblems_3 = load_npz_from_url_cached("https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/MotionPlanningProblems/fishbotics_mp_problems_part_3.npz")
        allproblems_4 = load_npz_from_url_cached("https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/MotionPlanningProblems/fishbotics_mp_problems_part_4.npz")
          
                        
        allproblems_1 = allproblems_1['arr_0'].item()
        allproblems_2 = allproblems_2['arr_0'].item()
        allproblems_3 = allproblems_3['arr_0'].item()
        allproblems_4 = allproblems_4['arr_0'].item()
        
        return {**allproblems_1, **allproblems_2, **allproblems_3, **allproblems_4}

    #######################################
    # Type check and conversion functions
    #######################################

    @staticmethod
    def is_a_number(obj):
        """
      Check if the argument is a float or int number
      
      Parameters
      ----------
      obj: object
          Object to be verified.
      
      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """

        return str(type(obj)) in ["<class 'int'>", "<class 'float'>", "<class 'numpy.float64'>"]

    @staticmethod
    def is_a_natural_number(obj):
        """
      Check if the argument is a natural number (integer and >=0)
      
      Parameters
      ----------
      obj: object
          Object to be verified.
      
      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """

        return str(type(obj)) == "<class 'int'>" and obj >= 0


        
    @staticmethod
    def is_a_matrix(obj, n=None, m=None):
        """
        Check if the argument is an n x m matrix of integers or floats.

        Parameters
        ----------
        obj: object
            Object to be verified.

        n: positive int
            Number of rows
            (default: it does not matter).

        m: positive int
            Number of columns
            (default: it does not matter).

        Returns
        -------
        is_type: boolean
            If the object is of the type.
        """

        # Check if obj is a numpy array or matrix
        if isinstance(obj, (np.ndarray, np.matrix)):
            # Check shape if dimensions are provided

            shp = np.shape(obj)
            rows = shp[0]
            cols = shp[1] if len(shp)>1 else  1

            if ((n is not None) and (rows != n)) or ((m is not None) and (cols !=m)):
                return False

            # Ensure all elements are integers or floats
            return np.issubdtype(obj.dtype, np.floating) or np.issubdtype(obj.dtype, np.integer)

        # Check if obj is a list (and therefore could represent a matrix)
        if isinstance(obj, list):
            try:
                return Utils.is_a_matrix(np.matrix(obj), n, m)
            except:
                return False
    
    @staticmethod
    def is_a_vector(obj, n=None):
        """
      Check if the argument is a vector of floats.
      A vector is very flexible. For example, a 3D vector (is_a_vector(x,3)) can be:
      x1 = [0,1,2]
      x2 = [[0],[1],[2]]
      x3 = (0,1,2)
      x4 = np.matrix([0,1,2])
      x5 = np.matrix([[0],[1],[2]])
      x6 = np.array([0,1,2])
      x7 = np.array([[0],[1],[2]])
            
      Parameters
      ----------
      obj: object
          Object to be verified.

      n: positive int
          Check if the vector is n-dimensional. If 'None', it does not matter.
          (default: 'None').

      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """
        try:
            p = Utils.cvt(obj)
            if n is None:
                return np.shape(p)[1] == 1
            else:
                return (np.shape(p)[0] == n) and (np.shape(p)[1] == 1)
        except:
            return False
        # return Utils.is_a_matrix(obj, n, 1) or Utils.is_a_matrix(obj, 1, n)

    @staticmethod
    def is_a_list_vector(obj, n=None):
        """
      Check if the argument is a list of vector of floats with the same dimension.

      Parameters
      ----------
      obj: object
          Object to be verified.

      n: positive int
          Check if all the vectors are n-dimensional. If 'None', it does not matter, 
          as long as they are equal.
          (default: 'None').

      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """
              
        if n is None:
            try:
                n_common = np.shape(Utils.cvt(obj[0]))[0]
            except:
                return False
        else:
            n_common = n
        
        for p in obj:
            if not Utils.is_a_vector(p, n_common):
                return False
            
        return True
    
    @staticmethod
    def is_a_pd_matrix(obj, n=None):
        """
      Check if the argument is a symmetric nxn positive (semi)-definite matrix.
      
      Parameters
      ----------
      obj: object
          Object to be verified.

      n: positive int
          Dimension of the square matrix
          (default: it does not matter).
    
      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """
        value = Utils.is_a_matrix(obj, n, n)

        if value:
            value = np.allclose(obj, np.transpose(obj), rtol=1e-05, atol=1e-08)

        if value:
            try:
                np.linalg.cholesky(obj)
            except:
                value = False

        return value

    @staticmethod
    def is_a_color(obj):
        """
      Check if the argument is a HTML-compatible string that represents a color.
      
      Parameters
      ----------
      obj: object
          Object to be verified.
      
      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """

        try:
            obj = obj.replace(" ", "")
            Color(obj)
            return True
        except:
            return False

    @staticmethod
    def is_a_name(string):
        """
      Check if the argument is a valid name for uaibot objects.
      Only characters [a-z], [A-z], [0-9] and '_' are allowed.
      However, variables should not begin with numbers.

      Parameters
      ----------
      string: string
          Name to be verified.

      Returns
      -------
      is_name: boolean
          If the name is a valid name.
      """

        if str(type(string)) == "<class 'str'>":
            if string=='':
                return True
            else:
                allowed1 = set(st.ascii_lowercase + st.ascii_uppercase + st.digits + "_")
                allowed2 = set(st.ascii_lowercase + st.ascii_uppercase + "_")

                return (set(string) <= allowed1) and (set(string[0]) <= allowed2)
        else:
            return False

    @staticmethod
    def get_uaibot_type(obj):
        """
      Return the UAIBot type of the object.
      Return the empty string if it is not an UAIBot object.
      
      Parameters
      ----------
      obj: object
          Object to be verified.
      
      Returns
      -------
      obj_type: string
          UAIBot type.
      """
        type_str = str(type(obj))

        ind = -1
        k = 0
        while ind == -1 and k < len(Utils.UAIBOT_NAME_TYPES):
            ind = type_str.find(Utils.UAIBOT_NAME_TYPES[k])
            k += 1

        if ind == -1:
            return ""
        else:
            ind1 = type_str.rfind('.')
            ind2 = type_str.rfind('>')
            return "uaibot." + type_str[ind1 + 1:ind2 - 1]

    @staticmethod
    def is_a_simple_object(obj):
        """
      Check if the argument is a simple object.
      Check the constant 'Utils.IS_SIMPLE' for a list of simple objects.

      Parameters
      ----------
      obj: object
          Object to be verified.

      Returns
      -------
      is_type: boolean
          If the object is of the type.
      """
        return Utils.get_uaibot_type(obj) in Utils.IS_SIMPLE

    @staticmethod
    def is_a_metric_object(obj):
        """
      Check if the argument is a metric object (if distance can be computed).
      Check the constant 'Utils.IS_METRIC' for a list of metric objects.

      Parameters
      ----------
      obj: object
          Object to be verified.

      Returns
      -------
      is_type: boolean
          If the object is of the type.
      """
        return Utils.get_uaibot_type(obj) in Utils.IS_METRIC
    
    @staticmethod
    def is_a_groupable_object(obj):
        """
      Check if the argument is a groupable object.
      Check the constant 'Utils.IS_GROUPABLE' for a list of groupable objects.

      Parameters
      ----------
      obj: object
          Object to be verified.
      
      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """
        return Utils.get_uaibot_type(obj) in Utils.IS_GROUPABLE

    @staticmethod
    def is_a_obj_sim(obj):
        """
      Check if the argument is an object that can be put into the simulator.
      Check the constant 'Utils.IS_OBJ_SIM' for a list of objects that can be put in the simulator.
      
      Parameters
      ----------
      obj: object
          Object to be verified.
      
      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """
        return Utils.get_uaibot_type(obj) in Utils.IS_OBJ_SIM

    @staticmethod
    def is_url_available(url, types):
        """
      Try to access the content of the url 'url'. Also verifies if the content is one of the extensions contained in
      'types' (e.g., types = ['png', 'bmp', 'jpg', 'jpeg'] for images).

      Never throws an Exception, always returning a string with a message. Returns 'ok!' if and only if the
      url was succesfully acessed and has the correct file type.

      Parameters
      ----------
      url: string
          The url string.

      types: list of string
          The desired content extensions.

      Returns
      -------
      message: string
          Message.
      """

        if not (str(type(url)) == "<class 'str'>"):
            return " is not a valid url"

        ind = url.rfind('.')
        filetype = url[ind + 1:]

        if not (filetype in types):
            return " must be an file of the types: '" + str(types) + "'"

        try:
            h = httplib2.Http(disable_ssl_certificate_validation=True) #Change this later
            resp = h.request(url, 'HEAD')
            if int(resp[0]['status']) < 400:
                return "ok!"
            else:
                return " : not able to access '" + url + "'."
        except:
            return " : not able to access '" + url + "'."

    @staticmethod
    def get_environment():
        try:
            import google.colab
            return "Colab"
        except ImportError:
            try:
                from IPython import get_ipython
                if "google.colab" in sys.modules:
                    return "Colab"
                elif get_ipython() is not None:
                    return "Jupyter"
            except ImportError:
                return "Not in Jupyter/Colab"
        return "Local"

    @staticmethod
    def obj_to_cpp(obj, htm=None):
            
            if Utils.get_uaibot_type(obj) != 'uaibot.PointCloud' and htm is None:
                htm = obj.htm
                
                
            if Utils.get_uaibot_type(obj) == 'uaibot.Ball':
                return ub_cpp.CPP_GeometricPrimitives.create_sphere(htm, obj.radius)
            if Utils.get_uaibot_type(obj) == 'uaibot.Box':
                return ub_cpp.CPP_GeometricPrimitives.create_box(htm, obj.width, obj.depth, obj.height)
            if Utils.get_uaibot_type(obj) == 'uaibot.Cylinder':
                return ub_cpp.CPP_GeometricPrimitives.create_cylinder(htm, obj.radius, obj.height)
            if Utils.get_uaibot_type(obj) == 'uaibot.PointCloud':
                return obj.cpp_pointcloud 
            if Utils.get_uaibot_type(obj) == 'uaibot.ConvexPolytope':
                Q = htm[0:3,0:3]
                p = htm[0:3,-1]
                A = obj.A * Q.transpose()
                b = obj.b + obj.A * Q.transpose()*p
                return ub_cpp.CPP_GeometricPrimitives.create_convexpolytope(htm, A, b)
                
    #######################################
    # Distance computation functions
    #######################################

    @staticmethod
    def softmin(x: List[float], h: float) -> float:
        minval = np.min(x)
        s=0

        for val in x:
            s+= exp(-h*(val-minval))

        return minval -(1/h)*np.log(s)

    @staticmethod
    def softselectmin(x: List[Vector], y: List[float], h: float) -> np.matrix:
        minval = np.min(x)
        s = 0

        coef = []
        for val in x:
            coef.append(exp(-(val - minval)/h))
            s += coef[-1]

        coef = [c/s for c in coef]

        sselect = 0 * y[0]
        for i in range(len(coef)):
            sselect += coef[i] * y[i]

        return sselect, minval - h * np.log(s/len(coef))


    @staticmethod
    def softmax(x: List[float], h: float) -> float:
        return Utils.softmin(x,-h)

    @staticmethod
    def softselectmax(x: List[Vector], y: List[float], h: float) -> np.matrix:
        return Utils.softselectmin(x,y,-h)

    @staticmethod
    def compute_aabbdist(obj1: MetricObject , obj2: MetricObject) -> float:

        box1 = obj1.aabb()
        box2 = obj2.aabb()
        delta = box1.htm[0:3, -1] - box2.htm[0:3, -1]

        dx = max(abs(delta[0, 0]) - (box1.width + box2.width) / 2, 0)
        dy = max(abs(delta[1, 0]) - (box1.depth + box2.depth) / 2, 0)
        dz = max(abs(delta[2, 0]) - (box1.height + box2.height) / 2, 0)

        return np.sqrt(dx ** 2 + dy ** 2 + dz ** 2)


    @staticmethod
    def _compute_dist_python(obj_a: MetricObject, obj_b: MetricObject, p_a: Vector, 
                            tol: float =0.001, no_iter_max: int =20) -> Tuple[np.matrix, np.matrix, float, List]:

        converged = False
        i = 0
        
        hist_error = []

        while (not converged) and i < no_iter_max:
            p_a_ant = p_a
            p_b, _ = obj_b.projection(p_a)
            p_a, _ = obj_a.projection(p_b)
            hist_error.append(np.linalg.norm(p_a - p_a_ant))
            converged = hist_error[-1] < tol
            i += 1

        dist = np.linalg.norm(p_a - p_b)

        return p_a, p_b, dist, []
                
    @staticmethod
    def compute_dist(obj_a: MetricObject, obj_b: MetricObject, p_a_init: Optional[Vector] = None, 
                     tol: float =0.001, no_iter_max: int=20, h: float=0, 
                     eps: float = 0, mode: str ='auto') -> Tuple[np.matrix, np.matrix, float, List]:
        """
    Compute Euclidean distance or differentiable distance between two objects.
    
    If h>0 or eps > 0, it computes the Euclidean distance and it uses GJK's algorithm.
    
    Else, it computes the differentiable distance through Generalized Alternating Projection (GAP).
    See the paper 'A Differentiable Distance Metric for Robotics Through Generalized Alternating Projection'.
    This only works in c++ mode, though.
    
    
    Parameters
    ----------
    obj_a : an object of type 'MetricObject' (see Utils.IS_METRIC)
        The first object.
        
    obj_b : an object of type 'MetricObject' (see Utils.IS_METRIC)
        The second object.
        
    p_a_init : a 3D vector (3-element list/tuple, (3,1)/(1,3)/(3,)-shaped numpy matrix/numpy array) or None
        Initial point for closest point in 'obj_a'. If 'None', is set to random.
        (default: None).
    
    tol : positive float
        Convergence criterion of GAP: it stops when ||a[k+1]-a[k]|| < tol.
        Only valid when h > 0 or eps > 0.
        (default: 0.001m).      

    no_iter_max : positive int 
        Maximum number of iterations of GAP.
        Only valid when h > 0 or eps > 0.
        (default: 20 iterations). 

    h : nonnegative float
        h parameter in the generalized distance function.
        If h=0 and eps=0, it is simply the Euclidean distance.
        (default: 0). 

    eps : nonnegative float
        h parameter in the generalized distance function.
        If h=0 and eps=0, it is simply the Euclidean distance.
        (default: 0). 

    mode : string
    'c++' for the c++ implementation, 'python' for the python implementation
    and 'auto' for automatic ('c++' is available, else 'python').
    (default: 'auto').
                                                    
    Returns
    -------
    point_a : 3 x 1 numpy matrix
        Closest point (Euclidean or differentiable) in obj_a.

    point_b : 3 x 1 numpy matrix
        Closest point (Euclidean or differentiable) in obj_b.

    distance : float
        Euclidean or differentiable distance.
        
    hist_error: list of floats
        History of convergence error.    
                
    """

        if mode=='python' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
            if Utils.get_uaibot_type(obj_a) == 'uaibot.PointCloud' or Utils.get_uaibot_type(obj_b) == 'uaibot.PointCloud':
                raise Exception("Point cloud distance is not supported in Python mode!")
            
            if not Utils.is_a_metric_object(obj_a):
                raise Exception("In python mode, the parameter 'obj_a' must be one of the following: " + str(Utils.IS_METRIC) + ".")

            if not Utils.is_a_metric_object(obj_b):
                raise Exception("In python mode, the parameter 'obj_b' must be one of the following: " + str(Utils.IS_METRIC) + ".")
            
        if mode=='c++':
            if not (Utils.get_uaibot_type(obj_a) == 'uaibot.CPP_GeometricPrimitives' or Utils.is_a_metric_object(obj_a)):
                raise Exception("In c++ mode, the parameter 'obj_a' must be 'uaibot.CPP_GeometricPrimitives' or "+ str(Utils.IS_METRIC) + ".")
            if not (Utils.get_uaibot_type(obj_b) == 'uaibot.CPP_GeometricPrimitives' or Utils.is_a_metric_object(obj_b)):
                raise Exception("In c++ mode, the parameter 'obj_b' must be 'uaibot.CPP_GeometricPrimitives' or "+ str(Utils.IS_METRIC) + ".")            


        if (mode == 'c++') or (mode=='auto' and os.environ['CPP_SO_FOUND']=='1'):
            if Utils.is_a_metric_object(obj_a):
                obj_a_cpp = Utils.obj_to_cpp(obj_a) 
            else:
                obj_a_cpp = obj_a

            if Utils.is_a_metric_object(obj_b):
                obj_b_cpp = Utils.obj_to_cpp(obj_b) 
            else:
                obj_b_cpp = obj_b
           
            
        if p_a_init is None:
            if ((mode == 'c++') or (mode=='auto' and os.environ['CPP_SO_FOUND']=='1')) and h >= 1e-5:
                dist_res = obj_a_cpp.dist_to(obj_b_cpp, 0.0, 0.0, tol, no_iter_max, np.matrix([0,0,0]).reshape((3,1)))
                p_a = Utils.cvt(dist_res.proj_A)
            else:
                p_a = Utils.cvt(np.random.uniform(-3, 3, size=(3,)))
        else:
            p_a = Utils.cvt(p_a_init)


        if not (p_a_init is None or Utils.is_a_vector(p_a_init, 3)):
            raise Exception("The optional parameter 'p_a_init' must be a 3D vector or 'None'.")

        if not Utils.is_a_number(tol) or tol <= 0:
            raise Exception("The optional parameter 'tol' must be a nonnegative number.")

        if not Utils.is_a_natural_number(no_iter_max):
            raise Exception("The optional parameter 'no_iter_max' must be a nonnegative integer.")
        
        if not Utils.is_a_number(h) or h < 0:
            raise Exception("The optional parameter 'h' must be a nonnegative number.")

        if not Utils.is_a_number(eps) or eps < 0:
            raise Exception("The optional parameter 'eps' must be a nonnegative number.")
                
        if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
            raise Exception("c++ mode is set, but .so file was not loaded!")
        # end error handling

        if mode == 'python' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
            if h >0 or eps > 0:
                raise Exception("In Python mode, smoothing parameters 'h' and 'eps' must be set to 0!")
            
            return Utils._compute_dist_python(obj_a, obj_b, p_a, tol, no_iter_max)
        else:
            dist_res = obj_a_cpp.dist_to(obj_b_cpp, h, eps, tol, no_iter_max, p_a)
            return Utils.cvt(dist_res.proj_A), Utils.cvt(dist_res.proj_B), dist_res.dist, dist_res.hist_error











from utils import *

from graphics.meshmaterial import *
from graphics.model3d import *

from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *

from .links import *


def _create_darwin_mini_leg_right(htm, name, color, opacity):
    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color):
        raise Exception("The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")

    link_info = [[0, 0, 0, 0],  # "theta" rotation in z
                 [0.055, 0, 0, -0.125],  # "d" translation in z
                 [np.pi / 2, 0, np.pi / 2, -np.pi / 2],  # "alfa" rotation in x
                 [0.035, 0.185, 0.171, 0],  # "a" translation in x
                 [0, 0, 0, 0]]  # joint type


    n = 4
    col_model = [[], [], [], []]

    # Create 3d objects

    Q01 = Utils.rotz(np.pi / 2) * Utils.rotz(link_info[0][0]) * Utils.trn([0, 0, link_info[1][0]]) * Utils.rotx(link_info[2][0]) * Utils.trn(
        [link_info[3][0], 0, 0])
    Q02 = Utils.rotz(link_info[0][1]) * Utils.trn([0, 0, link_info[1][1]]) * Utils.rotx(link_info[2][1]) * Utils.trn(
        [link_info[3][1], 0, 0])
    Q03 = Utils.rotz(link_info[0][2]) * Utils.trn([0, 0, link_info[1][2]]) * Utils.rotx(link_info[2][2]) * Utils.trn(
        [link_info[3][2], 0, 0])
    Q04 = Utils.rotz(np.pi / 2) * Utils.rotz(link_info[0][3]) * Utils.trn([0, 0, link_info[1][3]]) * Utils.rotx(link_info[2][3]) * Utils.trn(
        [link_info[3][3], 0, 0])

    base_3d_obj = []

    link_3d_obj = []

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPD_1.obj',
            0.004,
            Utils.inv_htm(Q01) * Utils.trn([-0.487, -0.315, -0.06]) * Utils.rotx(-np.pi / 2) * Utils.rotz(-0.3),
            MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/XL-320-2.obj',
                0.004,
                Utils.inv_htm(Q01) * Utils.trn([0.04, 0.73, 0.155]) * Utils.rotx(np.pi / 2) * Utils.trn(
                    [0.69, -0.09, 0.655]) * Utils.rotx(0.08) * Utils.roty(-3.14 / 2) * Utils.trn(
                    [-0.3, -0.179, 0.15]) * Utils.rotz(3.14 / 2) * Utils.rotx(3.14 / 2),
                MeshMaterial(metalness=0.8, clearcoat=0.5, roughness=0.5, normal_scale=[0.5, 0.5], color="#302b2b",
                             opacity=opacity, side="DoubleSide"))
        ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/frm_ext_02.obj',
            0.004,
            Utils.inv_htm(Q01 * Q02) * Utils.trn([0.317, -0.01 - 0.035, -0.435]) * Utils.rotx(0.275) * Utils.roty(
                -np.pi / 2),
            MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPO.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02) * Utils.trn([0.317, -0.35, -0.345]) * Utils.roty(-np.pi / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPO.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02) * Utils.trn([0.04, -0.01 + 0.925, 0.155]) * Utils.rotx(np.pi / 2) * Utils.trn(
                    [0.69, -0.09, 0.655]) * Utils.rotx(0.08) * Utils.roty(-3.14 / 2) * Utils.trn(
                    [-0.3, -0.179, 0.15]) * Utils.rotz(3.14 / 2) * Utils.rotx(3.14 / 2),
                MeshMaterial(metalness=0.8, clearcoat=0.5, roughness=0.5, normal_scale=[0.5, 0.5], color="#302b2b",
                             opacity=opacity, side="DoubleSide"))
        ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/mini_cr_calf.obj',
            0.004,
            Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([-0.271, 0.445, -0.325]),
            MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPO.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([0.317, -0.35 + 0.185, -0.345]) * Utils.roty(-np.pi / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/XL-320-2.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02) * Utils.trn([0.04, -0.01 + 0.925, 0.155]) * Utils.rotx(np.pi / 2) * Utils.trn(
                    [0.69, -0.09, 0.655]) * Utils.rotx(0.08) * Utils.roty(-3.14 / 2) * Utils.trn(
                    [-0.3, -0.179, 0.15]) * Utils.rotz(3.14 / 2) * Utils.rotx(3.14 / 2),
                MeshMaterial(metalness=0.8, clearcoat=0.5, roughness=0.5, normal_scale=[0.5, 0.5], color="#302b2b",
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPO.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02) * Utils.trn([0.317, -0.35 + 0.171, -0.345]) * Utils.roty(-np.pi / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide"))
        ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/XL-320-2.obj',
            0.004,
            Utils.inv_htm(Q01 * Q02 * Q03 * Q04) * Utils.trn([0, 0.785, 0]) * Utils.rotz(np.pi) * Utils.trn(
                [0.3, 0.295, -0.062]) * Utils.rotz(np.pi / 2) * Utils.roty(-np.pi / 2) * Utils.trn(
                [0.04, 1, 0]) * Utils.rotx(np.pi / 2) * Utils.trn(
                [0.69, -0.09, 0.655]) * Utils.rotx(0.08) * Utils.roty(-3.14 / 2) * Utils.trn(
                [-0.3, -0.179, 0.15]) * Utils.rotz(3.14 / 2) * Utils.rotx(3.14 / 2),
            MeshMaterial(metalness=0.8, clearcoat=0.5, roughness=0.5, normal_scale=[0.5, 0.5], color="#302b2b",
                         opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPD_1.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03 * Q04) * Utils.trn([-0.487, 0.05, -0.06]) * Utils.rotx(
                    -np.pi / 2) * Utils.rotz(-0.3),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/foot.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03 * Q04) * Utils.trn([-0.06 + 0.34, 0.19, -0.31]) * Utils.rotz(np.pi),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/sole.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03 * Q04) * Utils.trn([-0.06 + 0.34, 0.19, -0.31]) * Utils.rotz(np.pi),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide"))
        ]
    )

    # Create links

    links = []
    for i in range(n):
        links.append(Link(i, link_info[0][i], link_info[1][i], link_info[2][i], link_info[3][i], link_info[4][i],
                          link_3d_obj[i]))

        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration
    q0 = [0.000, 0.000, 0.000, 0.000]

    # Create joint limits
    joint_limits = (np.pi / 180) * np.matrix([[-180, 180], [-180, 180], [-180, 180], [-180, 180]])

    return base_3d_obj, links, np.identity(4), Utils.trn([-0.037,-0.025,-0.129]), q0, joint_limits



try:
    # New versions
    from IPython.display import display, HTML
except ImportError:
    # Fallback for old versions
    from IPython.core.display import display, HTML
import re
from utils import *
from graphics.meshmaterial import *
from simobjects.box import *
from simobjects.pointlight import *
from simobjects.pointcloud import *
from simobjects.frame import *
import time
from pathlib import Path
import httplib2
import sys
from uaibot.utils.types import HTMatrix, Matrix, Vector, SimObject
from typing import Optional, Tuple, List
import os
import inspect

class Simulation:
    """
  A simulation variable.

  Parameters
  ----------
  obj_list : A list of objects that can be simulated (see Utils.IS_OBJ_SIM)
      The objects that will be added initially to the simulation.
      (default: empty list)

  ambient_light_intensity : float
      The intensity of the ambient light.
      (default: 12).

  ldr_urls : a list of six url strings or None
      A list containing the LDR lightning images in the following order:
      [positive_x, negative_x, positive_y, negative_y, positive_z, negative_z].
      If None, no LDR is used.
      (default: None).

  camera_type : string
      The camera type, either "orthographic" or "perspective".
      (default: "perspective").

  width : positive float
      The canvas width, in pixels.
      (default: [] (automatic)).

  height : positive float
      The canvas height, in pixels.
      (default: [] (automatic)).

  show_world_frame: boolean
      If the frame in the middle of the scenario is shown.
      (default: True).

  show_grid : boolean
      If the grid in the scenario is shown.
      (default: True).

  load_screen_color : string, a HTML-compatible color
      The color of the loading screen.
      (default: "#19bd39").

  background_color : string, a HTML-compatible color
      The color of the background.
      (default: "white").

  camera_start_pose: vector or list with 7 entries, or None
      The camera starting configuration. The first three elements is the camera position (x,y,z).
      The next three is a point in which the camera is looking at.
      The final one is the camera zoom.
      If None, uses a default configuration for the camera.
      (default: None).
  """

    _CAMERATYPE = ['perspective', 'orthographic']
    # Import the javascript code as a string

    _STRJAVASCRIPT = "<html>\n"

    _STRJAVASCRIPT += "<style>\n"
    _STRJAVASCRIPT += ".controller:hover{opacity:1 !important;}\n"
    _STRJAVASCRIPT += "</style>\n"

    _STRJAVASCRIPT += "<body>\n"

    _STRJAVASCRIPT += "<div id='canvas_container_##SIMID##' style='width:##WIDTH##px;height:##HEIGHT##px;position:relative'>\n"
    _STRJAVASCRIPT += "<div id='loading_screen_##SIMID##' style='width:##WIDTH##px;height:##HEIGHT##px;position:relative; " \
                      "background-color: ##LOADSCREENCOLOR##;text-align:center;align-items:center;display:flex;justify-content:center'> \n "
    _STRJAVASCRIPT += "<img src='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/SVG/logo_uai_bot.svg' style='width:200px;height:114px'/>\n "
    _STRJAVASCRIPT += "</div>\n"
    _STRJAVASCRIPT += "<script id='MathJax-script' async src='https://cdn.jsdelivr.net/npm/mathjax@3.0.1/es5/tex-mml-chtml.js'></script>\n"
    _STRJAVASCRIPT += "<canvas id='scene_##SIMID##' width='##WIDTH##px' height='##HEIGHT##px'></canvas>\n"
    _STRJAVASCRIPT += "<!-- USER DIVS GO HERE -->"
    _STRJAVASCRIPT += "<div class = 'controller' style='width:##WIDTH##px;height:30px;'></div>\n"
    _STRJAVASCRIPT += "</div>\n"
    _STRJAVASCRIPT += "\n <script type=\"module\">\n"

    p = Path(__file__).with_name('threejs_sim.js')

    for line in open(p.absolute()).readlines():
        _STRJAVASCRIPT += line

    _STRJAVASCRIPT += "\n </script>"
    _STRJAVASCRIPT += "\n </body>"
    _STRJAVASCRIPT += "\n </html>"

    #######################################
    # Attributes
    #######################################

    @property
    def list_of_objects(self) -> List["SimObject"]:
        """A list of all sim objects."""
        return self._list_of_objects

    @property
    def list_of_names(self) -> List[str]:
        """A list of all object names."""
        return self._list_of_names

    @property
    def ambient_light_intensity(self) -> float:
        """Ambient light intensity."""
        return self._ambient_light_intensity

    @property
    def ldr_urls(self) -> List[str]:
        """A list of the LDR light urls."""
        return self._ldr_urls

    @property
    def camera_type(self) -> str:
        """Type of the camera."""
        return self._camera_type

    @property
    def width(self) -> int:
        """Width, in pixels, of the canvas"""
        return self._width

    @property
    def height(self) -> int:
        """Height, in pixels, of the canvas"""
        return self._height

    @property
    def show_world_frame(self) -> bool:
        """If the world frame is shown"""
        return self._show_world_frame

    @property
    def show_grid(self) -> bool:
        """If the grid in the world is shown"""
        return self._show_grid

    @property
    def load_screen_color(self) -> str:
        """Loading screen color"""
        return self._load_screen_color

    @property
    def background_color(self) -> str:
        """Color of the background of the scenario"""
        return self._background_color

    @property
    def camera_start_pose(self) -> List[float]:
        """The camera starting pose. The first three elements are the starting camera position, the next three ones
        is the starting point in which the camera is looking at and the last one is the zoom"""
        return self._camera_start_pose
    
    @property
    def pixel_ratio(self) -> float:
        """Relate to the resolution. Make this number smaller to ahve a performance boost"""
        return self._pixel_ratio

    @property
    def anti_aliasing(self) -> bool:
        """If the animation uses anti aliasing. Make this number smaller to ahve a performance boost"""
        return self._anti_aliasing
    
    #######################################
    # Constructor
    #######################################

    def __init__(self, obj_list: List["SimObject"]=[], ambient_light_intensity: float =4, 
                 ldr_urls: Optional[List[str]] = None, camera_type: str ="perspective", 
                 width: Optional[int] =[], height: Optional[int] = [], 
                 show_world_frame: bool = True, show_grid: bool = True, 
                 load_screen_color: str ="#19bd39", background_color: str ="#F5F6FA",
                 camera_start_pose: Optional[List[float]] = None,
                 pixel_ratio : float = 1.0, anti_aliasing: bool = True) -> "Simulation":

        if not Utils.is_a_number(ambient_light_intensity) or ambient_light_intensity < 0:
            raise Exception("The parameter 'ambient_light_intensity' should be a nonnegative float.")

        if not (camera_type in Simulation._CAMERATYPE):
            raise Exception("The parameter 'camera_type' must be one of the following strings: " + str(
                Simulation._CAMERATYPE) + ".")

        
        if (not width==[]) and (not Utils.is_a_number(width) or width <= 0):
            raise Exception("The parameter 'width' must be a positive float or [] (automatic).")

        if (not height==[]) and (not Utils.is_a_number(height) or height <= 0):
            raise Exception("The parameter 'height' must be a positive float or [] (automatic).")
        
        if not str(type(show_world_frame)) == "<class 'bool'>":
            raise Exception("The parameter 'show_world_frame' must be a boolean.")

        if not str(type(show_grid)) == "<class 'bool'>":
            raise Exception("The parameter 'show_grid' must be a boolean.")

        if not Utils.is_a_color(load_screen_color):
            raise Exception("The parameter 'load_screen_color' must be a HTML-compatible color.")

        if not Utils.is_a_color(background_color):
            raise Exception("The parameter 'background_color' must be a HTML-compatible color.")

        if not (ldr_urls is None):
            if not (str(type(ldr_urls)) == "<class 'list'>") or not (len(ldr_urls) == 6):
                raise Exception("The parameter 'ldr_urls' should be a list of six urls or 'None'.")
            else:
                for url in ldr_urls:
                    error = Utils.is_url_available(url, ['png', 'bmp', 'jpg', 'jpeg'])
                    if not (error == "ok!"):
                        raise Exception("The parameter 'url' #F5F6FA" + error)

        if camera_start_pose is None:
            if camera_type=="perspective":
                camera_start_pose = [1.76, 1.10, 1.45, 0, 0, 0, 1]
            else:
                camera_start_pose = [1.3, 1.8, 2.7, 0, 0, 0, 4]

        if not Utils.is_a_vector(camera_start_pose,7):
            raise Exception("The parameter 'camera_start_pose' should be either None or a 7 element vector.")

        if not Utils.is_a_number(pixel_ratio) or pixel_ratio < 0 or pixel_ratio>1:
            raise Exception("The parameter 'pixel_ratio' should be a float number between 0 and 1.")

        if not str(type(anti_aliasing)) == "<class 'bool'>":
            raise Exception("The parameter 'anti_aliasing' must be a boolean.")
                
        self._list_of_objects = []
        self._list_of_names = []
        self._ambient_light_intensity = ambient_light_intensity
        self._camera_type = camera_type
        self._ldr_urls = ldr_urls
        
        if width==[] or height==[]:
            if Utils.get_environment() == 'Local':
                self._width = 960 
                self._height = 720 
            else:
                self._width = 800
                self._height = 600                
        else:
            self._width = width
            self._height = height
            
        self._show_world_frame = show_world_frame
        self._show_grid = show_grid
        self._load_screen_color = load_screen_color
        self._background_color = background_color
        self._camera_start_pose = np.array(camera_start_pose).tolist()
        self._pixel_ratio = pixel_ratio
        self._anti_aliasing = anti_aliasing
   
             
        #Add reference frame
        if self._show_world_frame:
            self.add(Frame(name='w0_frame', htm=Utils.trn([0,0,0]),size=0.5))
        
        self.add(obj_list)


    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Simulation: \n\n"
        string += " Variables: \n"
        string += str(self.list_of_names)
        string += " Width: "+str(self.width)+"px, Height: "+str(self.height)+"px\n"
        string += " Camera type: "+str(self.camera_type)+"\n"
        string += " Ambient light intensity: " + str(self.ambient_light_intensity) + "\n"
        string += " Show world frame: " + str(self.show_world_frame) + "\n"
        string += " Show grid: " + str(self.show_grid) + "\n"
        string += " Background color: " + str(self.background_color) + "\n"

        return string

    #######################################
    # Methods
    #######################################

    @staticmethod
    def create_sim_factory(objects: List["SimObject"] =[], light_intensity: float = 1.0) -> "Simulation":
        """
    Create an environment of a factory.
    Factory panorama taken from:
    'https://www.samrohn.com/360-panorama/chrysler-factory-detroit-usa-360-tour/chrysler-factory-360-panorama-tour-007/'

    Parameters
    ----------
    objects: list of objects that can be simulated (see Utils.IS_OBJ_SIM)
        The objects to be added to the scenario.

    Returns
    -------
    sim: 'Simulation' object
        Simulation object.
    """

        texture_ground = Texture(
            url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/Textures/factory_ground.png',
            wrap_s='RepeatWrapping', wrap_t='RepeatWrapping', repeat=[4, 4])

        mesh_ground = MeshMaterial(texture_map=texture_ground, metalness=1, roughness=1)

        ground = Box(name="ground", width=6, depth=6, height=0.01, htm=Utils.trn([0, 0, -0.005]),
                     mesh_material=mesh_ground)

        light1 = PointLight(name="light1", color="white", intensity=light_intensity, htm=Utils.trn([-1,-1, 1.5]))
        light2 = PointLight(name="light2", color="white", intensity=light_intensity, htm=Utils.trn([-1, 1, 1.5]))
        light3 = PointLight(name="light3", color="white", intensity=light_intensity, htm=Utils.trn([ 1,-1, 1.5]))
        light4 = PointLight(name="light4", color="white", intensity=light_intensity, htm=Utils.trn([ 1, 1, 1.5]))

        ldr_url = "https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/LDR/factory_"
        ldr_list = [ldr_url + "px.png", ldr_url + "nx.png", ldr_url + "py.png", ldr_url + "ny.png", ldr_url + "nz.png",
                    ldr_url + "nz.png"]

        sim = Simulation(objects, ambient_light_intensity=2*light_intensity, ldr_urls=ldr_list)
        sim.add(ground)
        sim.add(light1)
        sim.add(light2)
        sim.add(light3)
        sim.add(light4)

        return sim
    
    @staticmethod
    def create_sim_mountain(objects: List["SimObject"] =[], light_intensity: float = 1.0) -> "Simulation":
        """
    Create an environment of a mountain.
    Outside panorama taken from:
    'https://opengameart.org/content/skiingpenguins-skybox-pack'

    Parameters
    ----------
    objects: list of objects that can be simulated (see Utils.IS_OBJ_SIM)
        The objects to be added to the scenario.

    Returns
    -------
    sim: 'Simulation' object
        Simulation object.
    """

        texture_ground = Texture(
            url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/Textures/grass.png',
            wrap_s='RepeatWrapping', wrap_t='RepeatWrapping', repeat=[100, 100])

        mesh_ground = MeshMaterial(texture_map=texture_ground, metalness=1, roughness=1)

        ground = Box(name="ground", width=100, depth=100, height=0.01, htm=Utils.trn([0, 0, -0.005]),
                     mesh_material=mesh_ground)

        light1 = PointLight(name="light1", color="white", intensity=light_intensity, htm=Utils.trn([-1,-1, 1.5]))
        light2 = PointLight(name="light2", color="white", intensity=light_intensity, htm=Utils.trn([-1, 1, 1.5]))
        light3 = PointLight(name="light3", color="white", intensity=light_intensity, htm=Utils.trn([ 1,-1, 1.5]))
        light4 = PointLight(name="light4", color="white", intensity=light_intensity, htm=Utils.trn([ 1, 1, 1.5]))

        ldr_url = "https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/LDR/green_"
        ldr_list = [ldr_url + "px.jpg", ldr_url + "nx.jpg", ldr_url + "py.jpg", ldr_url + "ny.jpg", ldr_url + "pz.jpg",
                    ldr_url + "nz.jpg"]


        sim = Simulation(objects, ambient_light_intensity=2*light_intensity, ldr_urls=ldr_list)
        sim.add(ground)
        sim.add(light1)
        sim.add(light2)
        sim.add(light3)
        sim.add(light4)

        return sim

    @staticmethod
    def create_sim_hill(objects: List["SimObject"] =[], light_intensity: float = 1.0) -> "Simulation":
        """
    Create an environment of a hill.
    Outside panorama taken from:
    'https://polyhaven.com/a/spaichingen_hill'

    Parameters
    ----------
    objects: list of objects that can be simulated (see Utils.IS_OBJ_SIM)
        The objects to be added to the scenario.

    Returns
    -------
    sim: 'Simulation' object
        Simulation object.
    """

        texture_ground = Texture(
            url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/Textures/factory_ground.png',
            wrap_s='RepeatWrapping', wrap_t='RepeatWrapping', repeat=[4, 4])


        # texture_ground = Texture(
        #     url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/Textures/grass.png',
        #     wrap_s='RepeatWrapping', wrap_t='RepeatWrapping', repeat=[20, 20])

        mesh_ground = MeshMaterial(texture_map=texture_ground, metalness=1, roughness=1)

        ground = Box(name="ground", width=20, depth=20, height=0.01, htm=Utils.trn([0, 0, -0.005]),
                     mesh_material=mesh_ground)

        light1 = PointLight(name="light1", color="white", intensity=light_intensity, htm=Utils.trn([-1,-1, 1.5]))
        light2 = PointLight(name="light2", color="white", intensity=light_intensity, htm=Utils.trn([-1, 1, 1.5]))
        light3 = PointLight(name="light3", color="white", intensity=light_intensity, htm=Utils.trn([ 1,-1, 1.5]))
        light4 = PointLight(name="light4", color="white", intensity=light_intensity, htm=Utils.trn([ 1, 1, 1.5]))

        ldr_url = "https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/LDR/hill_"
        ldr_list = [ldr_url + "px.png", ldr_url + "nx.png", ldr_url + "py.png", ldr_url + "ny.png", ldr_url + "pz.png",
                    ldr_url + "nz.png"]


        sim = Simulation(objects, ambient_light_intensity=2*light_intensity, ldr_urls=ldr_list)
        sim.add(ground)
        sim.add(light1)
        sim.add(light2)
        sim.add(light3)
        sim.add(light4)

        return sim

    @staticmethod
    def create_orchard_road(objects: List["SimObject"] =[], light_intensity: float = 1.0) -> "Simulation":
        """
    Create an environment of an orchard road.
    Outside panorama taken from:
    'https://polyhaven.com/a/citrus_orchard_road'

    Parameters
    ----------
    objects: list of objects that can be simulated (see Utils.IS_OBJ_SIM)
        The objects to be added to the scenario.

    Returns
    -------
    sim: 'Simulation' object
        Simulation object.
    """

        texture_ground = Texture(
            url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/Textures/ground_orchard.jpg',
            wrap_s='RepeatWrapping', wrap_t='RepeatWrapping', repeat=[5, 5])

        mesh_ground = MeshMaterial(texture_map=texture_ground, metalness=1, roughness=1)

        ground = Box(name="ground", width=20, depth=20, height=0.01, htm=Utils.trn([0, 0, -0.005]),
                     mesh_material=mesh_ground)

        light1 = PointLight(name="light1", color="white", intensity=light_intensity, htm=Utils.trn([-1,-1, 1.5]))
        light2 = PointLight(name="light2", color="white", intensity=light_intensity, htm=Utils.trn([-1, 1, 1.5]))
        light3 = PointLight(name="light3", color="white", intensity=light_intensity, htm=Utils.trn([ 1,-1, 1.5]))
        light4 = PointLight(name="light4", color="white", intensity=light_intensity, htm=Utils.trn([ 1, 1, 1.5]))

        ldr_url = "https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/LDR/orchard_road_"
        ldr_list = [ldr_url + "px.png", ldr_url + "nx.png", ldr_url + "py.png", ldr_url + "ny.png", ldr_url + "pz.png",
                    ldr_url + "nz.png"]


        sim = Simulation(objects, ambient_light_intensity=2*light_intensity, ldr_urls=ldr_list)
        sim.add(ground)
        sim.add(light1)
        sim.add(light2)
        sim.add(light3)
        sim.add(light4)

        return sim

    @staticmethod
    def create_sim_grid(objects: List["SimObject"] =[], light_intensity: float = 3.0) -> "Simulation":
        """
    Create an environment of a grid.

    Parameters
    ----------
    objects: list of objects that can be simulated (see Utils.IS_OBJ_SIM)
        The objects to be added to the scenario.

    Returns
    -------
    sim: 'Simulation' object
        Simulation object.
    """


        
        texture_ground = Texture(
            url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/Textures/grid_dark.png',
            wrap_s='RepeatWrapping', wrap_t='RepeatWrapping', repeat=[4, 4])
        
        mesh_ground = MeshMaterial(texture_map=texture_ground, metalness=1, roughness=1)
        
        size = 20

        ground = Box(width=size, depth=size, height=0.01, htm=Utils.trn([0, 0, -0.005]),
                     mesh_material=mesh_ground)

        wall1 = Box(width=0.005, depth=size, height=size, htm=Utils.trn([size/2, 0, size/2]),
                     mesh_material=mesh_ground)

        wall2 = Box(width=0.005, depth=size, height=size, htm=Utils.trn([-size/2, 0, size/2]),
                     mesh_material=mesh_ground)

        wall3 = Box(width=size, depth=0.005, height=size, htm=Utils.trn([0, size/2, size/2]),
                     mesh_material=mesh_ground)

        wall4 = Box(width=size, depth=0.005, height=size, htm=Utils.trn([0, -size/2, size/2]),
                     mesh_material=mesh_ground)
                        
        light1 = PointLight(name="light1", color="white", intensity=light_intensity, htm=Utils.trn([-2,-2, 5.0]))
        light2 = PointLight(name="light2", color="white", intensity=light_intensity, htm=Utils.trn([-2, 2, 5.0]))
        light3 = PointLight(name="light3", color="white", intensity=light_intensity, htm=Utils.trn([ 2,-2, 5.0]))
        light4 = PointLight(name="light4", color="white", intensity=light_intensity, htm=Utils.trn([ 2, 2, 5.0]))
        light5 = PointLight(name="light5", color="white", intensity=light_intensity, htm=Utils.trn([ 0,0,5]))

        
        ldr_url = "https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/LDR/"
        # ldr_list = [ldr_url + "grid_ldr_dark.png", ldr_url + "grid_ldr_dark.png", ldr_url + "grid_ldr_dark.png", ldr_url + "grid_ldr_dark.png", ldr_url + "grid_ldr_dark.png",
        #             ldr_url + "grid_ldr_dark.png"]
        # ldr_list = [ldr_url + "grid_ldr.png", ldr_url + "grid_ldr.png", ldr_url + "grid_ldr.png", ldr_url + "grid_ldr.png", ldr_url + "grid_ldr.png",
        #             ldr_url + "grid_ldr.png"]



        sim = Simulation(objects, ambient_light_intensity=light_intensity, show_grid=False)
        sim.add(ground)
        sim.add(wall1)
        sim.add(wall2)
        sim.add(wall3)
        sim.add(wall4)
        sim.add(light1)
        sim.add(light2)
        sim.add(light3)
        sim.add(light4)
        sim.add(light5)

        return sim
            
    @staticmethod
    def create_sim_lesson(objects: List["SimObject"] =[], light_intensity: float = 1.0) -> "Simulation":
        """
    Create an environment for embedding into the lessons slides.

    Parameters
    ----------
    objects: list of objects that can be simulated (see Utils.IS_OBJ_SIM)
        The objects to be added to the scenario.

    Returns
    -------
    sim: 'Simulation' object
        Simulation object.
    """


                            
        light1 = PointLight(name="light1", color="white", intensity=light_intensity, htm=Utils.trn([-1,-1, 1.0]))
        light2 = PointLight(name="light2", color="white", intensity=light_intensity, htm=Utils.trn([-1, 1, 1.0]))
        light3 = PointLight(name="light3", color="white", intensity=light_intensity, htm=Utils.trn([ 1,-1, 1.0]))
        light4 = PointLight(name="light4", color="white", intensity=light_intensity, htm=Utils.trn([ 1, 1, 1.0]))
        light5 = PointLight(name="light5", color="white", intensity=light_intensity, htm=Utils.trn([ 0,0,5]))


        sim = Simulation(objects, ambient_light_intensity=2*light_intensity, show_grid=False)
        sim.add(light1)
        sim.add(light2)
        sim.add(light3)
        sim.add(light4)
        sim.add(light5)

        return sim

    def set_parameters(self, ambient_light_intensity: Optional[float]= None, ldr_urls: Optional[str] =None, 
                       camera_type: Optional[str] =None, width: Optional[int] =None,
                 height: Optional[int] = None, show_world_frame: Optional[bool] = None, 
                 show_grid: Optional[bool] = None, load_screen_color: Optional[str] = None, 
                 background_color: Optional[str] = None, camera_start_pose: Optional[List[float]] = None,
                 pixel_ratio : Optional[float] = None, anti_aliasing: Optional[bool] = None) -> None:
        """
      Change the simulation parameters.

      Parameters
      ----------
      ambient_light_intensity : float
          The intensity of the ambient light.
          If None, does not change the current value.
          (default: None).

      ldr_urls : a list of six url strings or None
          A list containing the LDR lightning images in the following order:
          [positive_x, negative_x, positive_y, negative_y, positive_z, negative_z].
          If None, does not change the current value.
          (default: None).

      camera_type : string
          The camera type, either "orthographic" or "perspective".
          If None, does not change the current value.
          (default: None).

      width : positive float
          The canvas width, in pixels.
          If None, does not change the current value.
          (default: None).

      height : positive float
          The canvas height, in pixels.
          If None, does not change the current value.
          (default: None).

      show_world_frame: boolean
          If the frame in the middle of the scenario is shown.
          If None, does not change the current value.
          (default: None).

      show_grid : boolean
          If the grid in the scenario is shown.
          If None, does not change the current value.
          (default: None).

      load_screen_color : string, a HTML-compatible color
          The color of the loading screen.
          If None, does not change the current value.
          (default: None).

      background_color : string, a HTML-compatible color
          The color of the background.
          If None, does not change the current value.
          (default: None).

      camera_start_pose: vector or list with 7 entries, or None
          The camera starting configuration. The first three elements is the camera position (x,y,z).
          The next three is a point in which the camera is looking at.
          The final one is the camera zoom.
          If None, does not change the current value.
          (default: None).
          
      pixel_ratio: float between 0 and 1, or None
          A parameter that sets the resolution. Can be reduced from 1 to reduce the rendering load.
          (default: None).
          
      anti_aliasing: boolean, or None
          If anti_aliasing is used in rendering or not.
          (default: None).
      """

        if (not ambient_light_intensity is None) and (not Utils.is_a_number(ambient_light_intensity) or ambient_light_intensity < 0):
            raise Exception("The parameter 'ambient_light_intensity' should be a nonnegative float.")

        if (not camera_type is None) and  (not (camera_type in Simulation._CAMERATYPE) or (not ambient_light_intensity is None)):
            raise Exception("The parameter 'camera_type' must be one of the following strings: " + str(
                Simulation._CAMERATYPE) + ".")

        if (not width is None) and  (not Utils.is_a_number(width) or width <= 0):
            raise Exception("The parameter 'width' must be a positive float.")

        if (not height is None) and  (not Utils.is_a_number(height) or height <= 0):
            raise Exception("The parameter 'height' must be a positive float.")

        if (not show_world_frame is None) and (not str(type(show_world_frame)) == "<class 'bool'>"):
            raise Exception("The parameter 'show_world_frame' must be a boolean.")

        if (not show_grid is None) and (not str(type(show_grid)) == "<class 'bool'>"):
            raise Exception("The parameter 'show_grid' must be a boolean.")

        if (not load_screen_color is None) and (not Utils.is_a_color(load_screen_color)):
            raise Exception("The parameter 'load_screen_color' must be a HTML-compatible color.")

        if (not background_color is None) and  (not Utils.is_a_color(background_color)):
            raise Exception("The parameter 'background_color' must be a HTML-compatible color.")

        if not (ldr_urls is None):
            if not (str(type(ldr_urls)) == "<class 'list'>") or not (len(ldr_urls) == 6):
                raise Exception("The parameter 'ldr_urls' should be a list of six urls or 'None'.")
            else:
                for url in ldr_urls:
                    error = Utils.is_url_available(url, ['png', 'bmp', 'jpg', 'jpeg'])
                    if not (error == "ok!"):
                        raise Exception("The parameter 'url' " + error)


        if (not camera_start_pose is None) and (not Utils.is_a_vector(camera_start_pose,7)):
            raise Exception("The parameter 'camera_start_pose' should be either None or a 7 element vector.")

        if (not pixel_ratio is None) and (not Utils.is_a_number(pixel_ratio) or pixel_ratio < 0 or pixel_ratio>1):
            raise Exception("The parameter 'pixel_ratio' should be a float number between 0 and 1.")

        if (not anti_aliasing is None) and (not str(type(anti_aliasing)) == "<class 'bool'>"):
            raise Exception("The parameter 'anti_aliasing' must be a boolean.")
        
        if not ambient_light_intensity is None:
            self._ambient_light_intensity = ambient_light_intensity

        if not camera_type is None:
            self._camera_type = camera_type

        if not ldr_urls is None:
            self._ldr_urls = ldr_urls

        if not width is None:
            self._width = width

        if not height is None:
            self._height = height

        if not show_world_frame is None:
            if show_world_frame and (not self._show_world_frame):
                self.add(Frame(name='w0_frame', htm=Utils.trn([0,0,0]),size=0.5))
            if (not show_world_frame) and self._show_world_frame:
                for j in range(len(self._list_of_names)):
                    if self._list_of_names[j] == 'w0_frame':
                        j_remove=j
                    
                del self._list_of_names[j_remove]
                del self._list_of_objects[j_remove]
                    
            self._show_world_frame = show_world_frame

        if not show_grid is None:
            self._show_grid = show_grid

        if not load_screen_color is None:
            self._load_screen_color = load_screen_color

        if not background_color is None:
            self._background_color = background_color

        if not camera_start_pose is None:
            self._camera_start_pose = np.array(camera_start_pose).tolist()
           
        if not pixel_ratio is None: 
            self._pixel_ratio = pixel_ratio
            
        if not anti_aliasing is None:     
            self._anti_aliasing = anti_aliasing


        
    def gen_code(self):
        """Generate code for injection."""

        string = Simulation._STRJAVASCRIPT
        
        aa_value = 'true' if self.anti_aliasing else 'false'
        string = string.replace(
            'antialias: true',
            f'antialias: {aa_value}'
        )

        # Replace window.devicePixelRatio
        string = string.replace(
            'window.devicePixelRatio',
            f'window.devicePixelRatio*{self.pixel_ratio}'
        )
        

        for obj in self.list_of_objects:
            if Utils.get_uaibot_type(obj) == "uaibot.HTMLDiv":
                string = re.sub("<!-- USER DIVS GO HERE -->",
                                "<div id='" + obj.name + "'></div>\n <!-- USER DIVS GO HERE -->",
                                string)

        string = re.sub("//SIMULATION PARAMETERS GO HERE", "const delay = 500; \n //SIMULATION PARAMETERS GO HERE",
                        string)

        max_time = 0
        for obj in self.list_of_objects:
            string = re.sub("//USER INPUT GOES HERE", obj.gen_code(), string)
            max_time = max(max_time, obj._max_time)

        sim_id = str(time.time()).replace(".","")

        string = re.sub("//SIMULATION PARAMETERS GO HERE",
                        "const maxTime = " + str(max_time) + "; \n //SIMULATION PARAMETERS GO HERE",
                        string)
        string = re.sub("//SIMULATION PARAMETERS GO HERE",
                        "const cameraType= '" + self.camera_type + "'; \n //SIMULATION PARAMETERS GO HERE",
                        string)
        string = re.sub("//SIMULATION PARAMETERS GO HERE",
                        "const sceneID= '" + sim_id + "'; \n //SIMULATION PARAMETERS GO HERE",
                        string)
        string = re.sub("//SIMULATION PARAMETERS GO HERE",
                        "const showWorldFrame="+("true" if self.show_world_frame else "false")+"; \n //SIMULATION PARAMETERS GO HERE",
                        string)
        string = re.sub("//SIMULATION PARAMETERS GO HERE",
                        "const showGrid="+("true" if self.show_grid else "false")+"; \n //SIMULATION PARAMETERS GO HERE",
                        string)
        string = re.sub("//SIMULATION PARAMETERS GO HERE",
                        "const backgroundColor='"+self.background_color+"'; \n //SIMULATION PARAMETERS GO HERE",
                        string)
        string = re.sub("//SIMULATION PARAMETERS GO HERE",
                        "const ambientLightIntensity = " + str(self.ambient_light_intensity) + ";\n //SIMULATION PARAMETERS GO HERE",
                        string)
        string = re.sub("//SIMULATION PARAMETERS GO HERE",
                        "const ldrUrls = " + (
                            str(self.ldr_urls) if not (self.ldr_urls is None) else "[]") + ";\n //SIMULATION PARAMETERS GO HERE",
                        string)
        string = re.sub("//SIMULATION PARAMETERS GO HERE",
                        "const cameraStartPose = "+str(self.camera_start_pose)+";\n //SIMULATION PARAMETERS GO HERE",
                        string)
        string = re.sub("##WIDTH##", str(self.width), string)
        string = re.sub("##HEIGHT##", str(self.height), string)
        string = re.sub("##HEIGHTLOGO##", str(round(0.57*self.width)), string)
        string = re.sub("##LOADSCREENCOLOR##", self.load_screen_color, string)
        string = re.sub("##SIMID##", sim_id, string)


        return string

    def run(self) -> None:
        """Run simulation."""

        display(HTML(self.gen_code()))

    def save(self, address: Optional[str] = None, file_name : str = 'sim') -> None:
        """
    Save the simulation as a self-contained HTML file.

    Parameters
    ----------
    address : string
        The address of the path (example "D:\\"). Default is 'None', and saves
        to the folder that the script was run.
        (default: 'None').
    file_name: string
        The name of the file ("the .html" extension should not appear).
        (default: 'sim')

    """
        if (not (str(type(address)) == "<class 'str'>")) and (not address is None):
            raise Exception(
                "The parameter 'address' should be a string.")
        if not (str(type(file_name)) == "<class 'str'>"):
            raise Exception(
                "The parameter 'file_name' should be a string.")

        try:
            
            if address is None:
                frame = inspect.stack()[1] 
                caller_filepath = frame.filename
                current_folder = os.path.dirname(os.path.abspath(caller_filepath))
            else:
                current_folder = address

            file = open(current_folder + "/" + file_name + ".html", "w+")
            file.write(self.gen_code())
            file.close()
        except:
            raise Exception("Could not open the path '"+current_folder+"' and create the file '"+file_name+".html'.")

    def scan_group(self, group):
        for obj in group.list_of_objects:
            if Utils.get_uaibot_type(obj) == 'uaibot.Robot' and obj.eef_frame_visible:
                self.add(obj._eef_frame)
            if Utils.get_uaibot_type(obj) == 'uaibot.Group':
                self.scan_group(obj)
                
            
    def add(self, obj_sim: List["SimObject"]) -> None:
        """
    Add an object to the simulation. It should be an object that
    can be simulated (Utils.is_a_obj_sim(obj) is true).

    Parameters
    ----------
    obj_sim : object or list of object
        The object(s) to be added to simulation.
    """

        if str(type(obj_sim)) == "<class 'list'>":
            for obj in obj_sim:
                self.add(obj)
        else:
            # Error handling
            if not Utils.is_a_obj_sim(obj_sim):
                raise Exception("The parameter 'obj' should be one of the following: " + str(Utils.IS_OBJ_SIM) + ".")

            if obj_sim.name in self.list_of_names:
                raise Exception("The name '" + obj_sim.name + "' is already in the list of symbols.")

            # end error handling

            self._list_of_names.append(obj_sim.name)
            self._list_of_objects.append(obj_sim)
            
            if Utils.get_uaibot_type(obj_sim) == 'uaibot.Robot' and obj_sim.eef_frame_visible:
                self.add(obj_sim._eef_frame)
            if Utils.get_uaibot_type(obj_sim) == 'uaibot.Group':
                self.scan_group(obj_sim)



from utils import *
from graphics.meshmaterial import *
import numpy as np
from uaibot.utils.types import HTMatrix, Matrix, Vector, SimObject



class Model3D:
    """
   A class that represents a 3d model of an object.

   Parameters
   ----------
   url : string
      The url that contains the 3d object.
      It must have one of the following formats: 'obj', 'stl', 'dae'.

   scale : positive float
       The scaling parameter of the object.
       (default: 1).

   htm : 4x4 numpy matrix
       The htm of the 3d models. This is used to tune the 'default' htm for the object in the uaibot simulator.
       This is necessary because the 3d model can have a different 'default' pose than the desired one.
       (default: np.identity(4)).

   mesh_material : 'uaibot.MeshMaterial' object
       The mesh_material to be applied into the 3d model.
       (default: None).
   """

    #######################################
    # Attributes
    #######################################

    @property
    def url(self) -> str:
        """The 3d model url."""
        return self._url

    @property
    def scale(self) -> float:
        """The object scale."""
        return self._scale

    @property
    def htm(self) -> "HTMatrix":
        """Object pose. A 4x4 homogeneous transformation matrix written is scenario coordinates."""
        return self._htm

    @property
    def mesh_material(self) -> "MeshMaterial":
        """The model mesh material."""
        return self._mesh_material

    #######################################
    # Constructor
    #######################################

    def __init__(self, url: str ="", scale: float =1, htm: HTMatrix =np.identity(4), 
                 mesh_material: Optional["MeshMaterial"]=None) -> "Model3D":

        # Error handling
        
        ALLOWED_MESHES_TYPES = ["uaibot.MeshMaterial","uaibot.MTLMeshMaterial","uaibot.GLBMeshMaterial"] 


        error = Utils.is_url_available(url, ['obj','stl','dae','glb'])
        if not (error == "ok!"):
            raise Exception("The parameter 'url' " + error)

        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_number(scale) or scale < 0:
            raise Exception("The parameter 'scale' should be a float.")

        if not (Utils.get_uaibot_type(mesh_material) in ALLOWED_MESHES_TYPES or (mesh_material is None)):
            raise Exception("The parameter 'mesh_material' should be a "+str(ALLOWED_MESHES_TYPES)+" object or 'None'.")

        self._type = url[url.rfind(".")+1:len(url)+1]
        
        if (not self._type == 'obj') and  Utils.get_uaibot_type(mesh_material) == "uaibot.MTLMeshMaterial":
            raise Exception("The parameter 'mesh_material' is a of MTL type, but the object is not of type '.obj'. MTL files are only supported to obj.")

        if (not self._type == 'glb') and  Utils.get_uaibot_type(mesh_material) == "uaibot.GLBMeshMaterial":
            raise Exception("The parameter 'mesh_material' is a GLB type, but the object is not of type '.glb'.")


        # end error handling

        self._url = url
        self._scale = scale
        self._htm = htm
        


        if mesh_material is None:
            self._mesh_material = MeshMaterial()
        else:
            self._mesh_material = mesh_material

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "3D object model with url: '" + self.url + "': \n\n"

        return string

    #######################################
    # Methods
    #######################################

    def gen_code(self, name):
        """Generate code for injection."""

        string = self.mesh_material.gen_code(name)
        string += "const htm_" + name + " = new Matrix4(); \n"
        string += "htm_" + name + ".set(" \
                  + str(self.htm[0,0]) + "," + str(self.htm[0,1]) + "," + str(self.htm[0,2]) + "," + str(
            self.htm[0,3]) + "," \
                  + str(self.htm[1,0]) + "," + str(self.htm[1,1]) + "," + str(self.htm[1,2]) + "," + str(
            self.htm[1,3]) + "," \
                  + str(self.htm[2,0]) + "," + str(self.htm[2,1]) + "," + str(self.htm[2,2]) + "," + str(
            self.htm[2,3]) + "," \
                  + str(self.htm[3,0]) + "," + str(self.htm[3,1]) + "," + str(self.htm[3,2]) + "," + str(
            self.htm[3,3]) + ");\n"

        string += "const object3d_" + name + "={\n"
        string += "url: '" + self.url + "',\n"
        string += "scale: " + str(self.scale) + ",\n"
        string += "matrix: htm_" + name + ",\n"
        string += "type: '" + self._type + "',\n"
        string += "mesh_material: material_" + name + "};\n\n"

        return string



from utils import *

from graphics.meshmaterial import *
from graphics.model3d import *

from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *

from .links import *


def _create_epson_t6(htm, name, color, opacity):
    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color):
        raise Exception("The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")

    link_info = [[0, 0, 0], #"theta" rotation in z
                 [0, 0.000, 0],  # "d" translation in z
                 [0, 3.14, 0],  # "alfa" rotation in x
                 [1.3 / 4, 1.1 / 4, 0],  # "a" translation in x
                 [0, 0, 1]] #joint type

    n = 3

    # Collision model
    col_model = [[], [], []]

    col_model[0].append(Box(htm=Utils.trn([-0.15, 0, 0.03]),
                            name=name + "_C0_0", width=0.375, height=0.05, depth=0.1, color="red", opacity=0.3))

    col_model[1].append(Box(htm=Utils.trn([-0.14, 0, -0.15]),
                            name=name + "_C1_0", width=0.38, height=0.2, depth=0.143, color="green", opacity=0.3))

    col_model[2].append(Cylinder(htm=Utils.trn([0, 0, -0.2]),
                                 name=name + "_C2_0", radius=0.03, height=0.4, color="blue", opacity=0.3))

    # Create 3d objects
    htm1 = np.matrix([[1., 0., 0., 0.], [0., 0.0008, -1., 0.], [0., 1., 0.0008, 0.], [0., 0., 0., 1.]])
    htm2 = np.matrix([[1., 0., 0., -0.325], [0., 0.0008, -1., 0.], [0., 1., 0.0008, 0.], [0., 0., 0., 1.]])
    htm3 = np.matrix([[0., -0.0008, 1., 0.], [1., 0.0008, 0., 0.], [-0.0008, 1., 0.0008, 0.22], [0., 0., 0., 1.]])
    htm4 = np.matrix([[1., 0., 0., 0.], [0., 0.0008, 1., 0.], [0., -1., 0.0008, -0.06], [0., 0., 0., 1.]])
    htm5 = np.matrix([[1., 0., 0., 0.], [0., 0.0008, 1., 0.], [0., -1., 0.0008, -0.4], [0., 0., 0., 1.]])
    
    

    base_3d_obj = [
        Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/EpsonT6/Base.obj',
                0.001,
                htm1,
                MeshMaterial(metalness=0.3, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5],
                             color=color, opacity=opacity))]

    link_3d_obj = []

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/EpsonT6/T6Axis1.obj',
                 0.001,
                 htm2,
                 MeshMaterial(metalness=0.3, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                              opacity=opacity)),
         Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/EpsonT6/T6Cable.obj',
                 0.001,
                 htm3,
                 MeshMaterial(color="black", opacity=opacity))
         ]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/EpsonT6/T6Axis2.obj',
                 0.001,
                 htm4,
                 MeshMaterial(metalness=0.3, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/EpsonT6/T6Axis3.obj',
                 0.001,
                 htm5,
                 MeshMaterial(metalness=0.3, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                              opacity=opacity))]
    )

    # Create links

    links = []
    for i in range(n):
        links.append(Link(i, link_info[0][i], link_info[1][i], link_info[2][i], link_info[3][i], link_info[4][i],
                          link_3d_obj[i]))

        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration

    htm_base_0 = Utils.trn([0.25 / 4, 0, 0.8 / 4])

    q0 = [0.0, 0.0, 0.0]

    #Create joint limits
    c = (np.pi/180)
    joint_limits = np.matrix([[-c*132,c*132],[-c*141,c*141],[0+0.02,0.15+0.02]])

    
    return base_3d_obj, links, htm_base_0, np.identity(4), q0, joint_limits



from utils import *
import numpy as np
import os
from ._check_free_config import *



def _ikm(self, htm_tg, htm = None, q0=None, p_tol=0.005, a_tol=5, no_iter_max=2000, ignore_orientation=False, 
         no_tries = 40, check_joint = True, check_auto = True, obstacles = [], mode='auto'):
    
    n = len(self._links)
    if htm is None:
        htm = self.htm


        
    # Error handling
    if not Utils.is_a_matrix(htm_tg, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix or 'None'.")

    if not Utils.is_a_natural_number(no_iter_max):
        raise Exception("The parameter 'no_iter_max' should be a nonnegative integer number.")

    if not (q0 is None) and (not Utils.is_a_vector(q0, n)):
        raise Exception("The parameter 'q0' should be a " + str(n) + " dimensional vector or 'None'.")

    if (not Utils.is_a_number(p_tol)) or p_tol <= 0:
        raise Exception("The parameter 'p_tol' should be a nonnegative number.")

    if (not Utils.is_a_number(a_tol)) or a_tol <= 0:
        raise Exception("The parameter 'a_tol' should be a nonnegative number.")

    if not str(type(check_joint)) == "<class 'bool'>":
        raise Exception("The parameter 'check_joint' should be a boolean.")

    if not str(type(check_joint)) == "<class 'bool'>":
        raise Exception("The parameter 'check_auto' should be a boolean.")
        
    if not str(type(ignore_orientation)) == "<class 'bool'>":
        raise Exception("The parameter 'ignore_orientation' should be a boolean.")

    if not Utils.is_a_natural_number(no_tries):
        raise Exception("The parameter 'no_tries' should be a nonnegative integer number.")

    for obs in obstacles:
        if not Utils.is_a_metric_object(obs):
            raise Exception("The parameter 'list' should be a list of metric objects : " + str(Utils.IS_METRIC) + ".")
            
    if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
        raise Exception("c++ mode is set, but .so file was not loaded!")
        # end error handling
    
    for k in range(no_tries):
        
        if q0 is None:
            q0_mod = np.matrix(self.q)
            for i in range(n):
                q0_mod[i,0] = self.joint_limit[i,0]+np.random.rand()*(self.joint_limit[i,1]-self.joint_limit[i,0])
        else:
            q0_mod = np.matrix(q0)
            for i in range(n):
                q0_mod[i,0] += k*0.1*np.random.randn()
                q0_mod[i,0] = min(q0_mod[i,0], self.joint_limit[i,1])
                q0_mod[i,0] = max(q0_mod[i,0], self.joint_limit[i,0])
                
        q, success = _ikm_aux(self, htm_tg, htm, q0_mod, p_tol, a_tol, no_iter_max, ignore_orientation, mode)            
        
        if success: 
            ok, _, _ = self.check_free_config(Utils.cvt(q), htm, obstacles,
                                check_joint, check_auto, mode = mode)
            if ok:
                return Utils.cvt(q) 
 
    
    #Failed to find a solution
    raise Exception("Solution for IK not found. You can try the following: \n" \
                    " Increasing the maximum number of iterations, 'no_iter_max' (currently " + str(
        no_iter_max) + ")\n Increasing the maximum number of tries 'no_tries' (currently " + str(
        no_tries) + ")\n" \
                    " Increasing the tolerance for the position, 'p_tol' (currently " + str(p_tol) + " meters)\n" \
                                                                                                        " Increasing the tolerance for the orientation, 'a_tol' (currently " + str(
        a_tol) + " degrees).")
        



def _ikm_aux(self, htm_tg, htm = None, q0=None, p_tol=0.005, a_tol=5, no_iter_max=2000, ignore_orientation=False, mode='auto'):

    if mode == 'python' or  (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
        return _ikm_python(self, htm_tg, htm, Utils.cvt(q0), p_tol, a_tol, no_iter_max, ignore_orientation)
    else:
        out =  self.cpp_robot.ik(htm_tg, htm, Utils.cvt(q0), p_tol, a_tol, no_iter_max, ignore_orientation)
        return out.qf, out.success




# (Private function) Used in IK
def _evolve_config(self, q, htm, p_tol, a_tol, htm_tg, iter_remain, ignore_orientation):

    found = False
    zero_u = False
    iter_end = False

    eta = 0.03
    eps = 0.001
    tol_u = 0.00001 
    i = 0

    while (not found) and (not zero_u) and (not iter_end):

        r, jac_r = self.task_function(htm=htm, htm_tg = htm_tg, q = q, mode='python')
        
        # print(r)
        # print(jac_r)

        if ignore_orientation:
            r = r[0:3]
            jac_r = jac_r[0:3, :]
        else:
            r[3:6] = np.sqrt(r[3:6])

        r[0] = np.sign(r[0])*np.sqrt(np.abs(r[0]))
        r[1] = np.sign(r[1])*np.sqrt(np.abs(r[1]))
        r[2] = np.sign(r[2])*np.sqrt(np.abs(r[2]))
            
        u = Utils.dp_inv_solve(jac_r, -eta * r, eps, mode='python')
        q = q + u 
        


        e_pos = max([r[0]**2, r[1]**2, r[2]**2])

        if ignore_orientation:
            e_ori = 0
        else:
            e_ori = max([(180 / np.pi) * acos(min(max(1 - num * num, -1), 1)) for num in r[3:6]])

        i += 1
        
        found = (e_pos < p_tol) and (e_ori < a_tol)
        zero_u = max(abs(u)) < tol_u
        iter_end = i > iter_remain

    return found, i, q


# Inverse kinematics for the end-effector
def _ikm_python(self, htm_tg, htm, q0=None, p_tol=0.005, a_tol=5, no_iter_max=2000, ignore_orientation=False):
    n = len(self._links)

    j = 0
    found = False
    q = Utils.cvt(q0)
    no_iter_remain = no_iter_max

    while not found and no_iter_remain >= 0:
        found, i, q = _evolve_config(self, q, htm, p_tol, a_tol, htm_tg, no_iter_remain, ignore_orientation)
        no_iter_remain -= i
        if not found:
            q = np.matrix(np.reshape((2 * np.pi) * np.random.rand(n), (n, 1)))

    return q, found



from utils import *
import numpy as np

def _add_ani_frame(self, time, q=None, htm=None, enforce_joint_limits=False):

    n = len(self.links)

    if q is None:
        q = self.q

    if htm is None:
        htm = self.htm

    # Error handling
    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    if not Utils.is_a_number(time) or time < 0:
        raise Exception("The parameter 'time' should be a nonnegative float.")

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix")

    if not str(type(enforce_joint_limits)) == "<class 'bool'>":
        raise Exception("The parameter 'enforce_joint_limits' must be a boolean.")

    # end error handling
    self._q = Utils.cvt(q)

    if enforce_joint_limits:
        for i in range(len(self.links)):
            self._q[i,0] = min(max(self._q[i,0],self.joint_limit[i,0]),self.joint_limit[i,1])

    f = [time, htm[0,0].item(), htm[0,1].item(), htm[0,2].item(), htm[0,3].item(),
         htm[1,0].item(), htm[1,1].item(), htm[1,2].item(), htm[1,3].item(),
         htm[2,0].item(), htm[2,1].item(), htm[2,2].item(), htm[2,3].item(),
         0, 0, 0, 1, np.array(q).reshape((n,)).tolist()]

    self._htm = htm
    self._frames.append(f)
    self._max_time = max(self._max_time, time)
    
    if self.eef_frame_visible:
        self._eef_frame.add_ani_frame(time,self.fkm(q))

    # Update attached objects:
    if len(self.attached_objects) > 0:
        htm = self.fkm(q, 'eef', htm)
        for obj in self.attached_objects:
            obj[0].add_ani_frame(time, htm * obj[1])



from utils import *
import numpy as np
import os

def _fkm(self, q=None, axis='eef', htm=None, mode='auto'):

    if q is None:
        q = self.q
    if htm is None:
        htm = np.matrix(self.htm)

    n = len(self._links)

    # Error handling
    if mode not in ['python','c++','auto']:
        raise Exception("The parameter 'mode' should be 'python,'c++', or 'auto'.")
    
    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    if not (axis == "eef" or axis == "dh" or axis == "com"):
        raise Exception("The parameter 'axis' should be one of the following strings:\n" \
                        "'eef': End-effector \n" \
                        "'dh': All " + str(n) + " axis of Denavit-Hartenberg\n" \
                                                "'com': All " + str(
            n) + " axis centered at the center of mass of the objects.")

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")
    
    if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
        raise Exception("c++ mode is set, but .so file was not loaded!")
    # end error handling

    if mode == 'python' or axis == 'com' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
        return _fkm_python(self, Utils.cvt(q), axis, htm)
    else:
        fk_res = self.cpp_robot.fk(Utils.cvt(q), htm, False)
        if axis=='eef':
            return Utils.cvt(fk_res.htm_ee)
        else:
            return [Utils.cvt(m) for m in fk_res.htm_dh]

def _fkm_python(self, q, axis, htm):

    n = len(self._links)
    htm_dh = [np.matrix(np.zeros((4,4))) for i in range(n)]

    for i in range(n):
        if i == 0:
            htm_dh[i][:, :] = htm * self._htm_base_0
        else:
            htm_dh[i][:, :] = htm_dh[i - 1][:, :]

        if self.links[i].joint_type == 0:
            htm_dh[i][:, :] = htm_dh[i][:, :] * Utils.rotz(q[i,0])
        else:
            htm_dh[i][:, :] = htm_dh[i][:, :] * Utils.rotz(self._links[i].theta)

        if self.links[i].joint_type == 1:
            htm_dh[i][:, :] = htm_dh[i][:, :] * Utils.trn([0, 0, q[i,0]])
        else:
            htm_dh[i][:, :] = htm_dh[i][:, :] * Utils.trn([0, 0, self._links[i].d])

        htm_dh[i][:, :] = htm_dh[i][:, :] * Utils.rotx(self._links[i].alpha)
        htm_dh[i][:, :] = htm_dh[i][:, :] * Utils.trn([self._links[i].a, 0, 0])

    if axis == 'com':
        for i in range(n):
            htm_dh[i][0:3, 3] = htm_dh[i][0:3, 3] + htm_dh[i][0:3, 0:3] * self._links[i].com_coordinates

    if axis == 'eef':
        htm_dh = htm_dh[-1][:, :] * self.htm_n_eef

    return htm_dh



from utils import *


def _add_col_object(self, sim):
    if not Utils.get_uaibot_type(sim) == "uaibot.Simulation":
        raise Exception("The parameter 'sim' should be a 'uaibot.Simulation' object.")

    for i in range(len(self.links)):
        for j in range(len(self.links[i].col_objects)):
            sim.add(self.links[i].col_objects[j][0])



from ._constrained_control_demo_1 import _constrained_control_demo_1
from ._control_demo_1 import _control_demo_1



class Demo:
    
    @staticmethod
    def constrained_control_demo_1():
        return _constrained_control_demo_1()
    
    @staticmethod
    def control_demo_1():
        return _control_demo_1()



from typing import Union, List, TypeAlias, Tuple
import numpy as np
import numpy.typing as npt

#Union[List[float], npt.NDArray[np.float64], np.matrix]
#means that we can send either a list, a numpy float, of a numpy matrix

Vector: TypeAlias = Union[List[float], npt.NDArray[np.float64], np.matrix]
Matrix: TypeAlias = Union[npt.NDArray[np.float64], np.matrix]

class Utils:
    @staticmethod
    def S(v: Vector) -> np.matrix: ...
    
    @staticmethod
    def rot(axis: Vector, angle: float) -> np.matrix: ...
    
    @staticmethod
    def trn(vector: Vector) -> np.matrix: ...
    
    @staticmethod
    def rotx(angle: float) -> np.matrix: ...
    
    @staticmethod
    def roty(angle: float) -> np.matrix: ...

    @staticmethod
    def rotz(angle: float) -> np.matrix: ...

    @staticmethod
    def htm_rand(trn_min: float, trn_max: float, rot: float) -> np.matrix: ...
    
    @staticmethod
    def inv_htm(htm: float, trn_max: float, rot: float) -> np.matrix: ...
    
    @staticmethod
    def axis_angle(htm: Matrix) -> Tuple[Vector, float]: ...
    
    @staticmethod
    def euler_angles(htm: Matrix) -> Tuple[float, float, float]: ...
    
    @staticmethod
    def dp_inv(A: Matrix, eps: float) -> Matrix: ...
    
    @staticmethod
    def dp_inv_solve(A: Matrix, b: Vector, eps: float, mode: str) -> Vector: ...
    
    #Continue declaring the methods...


import numpy as np
from uaibot.utils.types import HTMatrix, Matrix, Vector, MetricObject
from typing import Optional, Tuple, List
from .robot import *

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .robot import Robot
    
class DistStructLinkLink:

    @property
    def link_number_1(self) -> int:
        """The index of the first link in which the collision point is."""
        return self._link_number_1

    @property
    def link_number_2(self) -> int:
        """The index of the second link in which the collision point is."""
        return self._link_number_2

    @property
    def link_col_obj_number_1(self) -> int:
        """The index of the collision object of the first link in which the collision point is."""
        return self._link_col_obj_number_1

    @property
    def link_col_obj_number_2(self) -> int:
        """The index of the collision object of the second link in which the collision point is."""
        return self._link_col_obj_number_2

    @property
    def distance(self) -> float:
        """The distance (in meters) between the two links."""
        return self._distance

    @property
    def point_link_1(self) -> np.matrix:
        """The closest point (witness) in the first link. Written in scenario coordinates and in meters."""
        return np.matrix(self._point_link_1)

    @property
    def point_link_2(self) -> np.matrix:
        """The closest point (witness) in the second link. Written in scenario coordinates and in meters."""
        return np.matrix(self._point_link_2)

    @property
    def jac_distance(self) -> np.matrix:
        """The Jacobian of the distance in the robot's configuration space."""
        return np.matrix(self._jac_distance)

    #######################################
    # Constructor
    #######################################

    def __init__(self, link_number_1: int, link_col_obj_number_1: int, link_number_2: int, 
                 link_col_obj_number_2: int, distance: float, point_link_1: np.matrix, 
                 point_link_2: np.matrix, jac_distance: np.matrix) -> "DistStructLinkLink":

        self._link_number_1 = link_number_1
        self._link_number_2 = link_number_2
        self._link_col_obj_number_1 = link_col_obj_number_1
        self._link_col_obj_number_2 = link_col_obj_number_2
        self._distance = distance
        self._point_link_1 = point_link_1
        self._point_link_2 = point_link_2
        self._jac_distance = jac_distance

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string  = "Distance info: \n"
        string += " First Link number: "+str(self.link_number_1)+"\n"
        string += " Collision object of the first link number: "+str(self.link_col_obj_number_1 )+"\n"
        string += " Second Link number: "+str(self.link_number_2)+"\n"
        string += " Collision object of the second link number: "+str(self.link_col_obj_number_2 )+"\n"
        string += " Distance: " + str(self.distance) + " m\n"
        string += " Point in the first link: " + str(self.point_link_1.tolist()) + " m\n"
        string += " Point in the second link: " + str(self.point_link_2.tolist()) + " m\n"
        string += " Jacobian distance: " + str(self._jac_distance.tolist()) + "\n"
        return string


class DistStructRobotAuto:

    #######################################
    # Attributes
    #######################################

    @property
    def robot(self) -> "Robot":
        """Return the associated robot."""
        return self._robot

    @property
    def jac_dist_mat(self) -> np.matrix:
        """
		Return the matrix in which each row we have the distance Jacobian (gradient) for each robot link.
		"""
        return np.matrix(self._jac_dist_mat)

    @property
    def dist_vect(self) -> np.matrix:
        """
		Return the column vector in which each row we have the distance for each robot link.
		"""
        return np.matrix(self._dist_vect).reshape((self.no_items, 1))

    @property
    def no_items(self) -> int:
        """Return the number of items."""
        return self._no_items

    def __getitem__(self, key):
        return self._list_info[key]

    #######################################
    # Constructor
    #######################################

    def __init__(self, robot: "Robot") -> "DistStructRobotAuto":

        self._robot = robot
        self._robot_name = robot.name
        self._no_items = 0

        n = len(robot.links)
        self._list_info = []
        self._jac_dist_mat = np.matrix(np.zeros((0, n)))
        self._dist_vect = []

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        return "Distance struct between robot '" + self.robot.name + "' and itself (auto collision) '" \
               " with " + str(self.no_items) + " items"

    #######################################
    # Methods
    #######################################

    def _append(self, link_number_1, link_col_obj_number_1, link_number_2, 
                link_col_obj_number_2, distance, point_link_1, point_link_2, jac_distance) :

        self._list_info.append(
            DistStructLinkLink(link_number_1, link_col_obj_number_1, link_number_2, link_col_obj_number_2,
                               distance, point_link_1, point_link_2, jac_distance))

        self._jac_dist_mat = np.vstack((self._jac_dist_mat, jac_distance))
        self._dist_vect.append(distance)
        self._no_items += 1


    def get_item(self, link_number_1: int, link_col_obj_number_1: int, 
                 link_number_2: int, link_col_obj_number_2: int) -> "DistStructLinkLink":
        for dist_info in self._list_info:
            if link_number_1 == dist_info.link_number_1 and link_col_obj_number_1 == dist_info.link_col_obj_number_1 \
            and link_number_2 == dist_info.link_number_2 and link_col_obj_number_2 == dist_info.link_col_obj_number_2:
                return dist_info

        raise Exception("Item not found!")

    def get_closest_item(self) -> "DistStructLinkLink":
        minimum_distance = float('inf')
        index_minimum = -1
        for i in range(self._no_items):
            if self[i].distance < minimum_distance:
                minimum_distance = self[i].distance
                index_minimum = i

        return self[index_minimum]




from utils import *
import numpy as np
import math
from ._dist_struct_robot_obj import DistStructRobotObj
from ._dist_struct_robot_obj import DistStructLinkObj
import os
if os.environ['CPP_SO_FOUND']=="1":
    import uaibot_cpp_bind as ub_cpp

# Compute the distance from each link to an object, for the current configuration
# of the robot


def _diststructlinkobj_cpp2py(dslo_cpp):

    dslo_py = DistStructLinkObj(dslo_cpp.link_number, dslo_cpp.link_col_obj_number, dslo_cpp.distance, np.matrix(dslo_cpp.point_link).reshape((3,1)), np.matrix(dslo_cpp.point_object).reshape((3,1)), np.matrix(dslo_cpp.jac_distance))

    #dslo_py = DistStructLinkObj(dslo_cpp.link_number, dslo_cpp.link_col_obj_number, dslo_cpp.distance, dslo_cpp.point_link, dslo_cpp.point_object, dslo_cpp.jac_distance)


    return dslo_py

def _diststructlinkobj_py2cpp(dslo_py):

    dslo_cpp = ub_cpp.CPP_DistStructLinkObj()

    dslo_cpp.is_null = False
    dslo_cpp.link_number = dslo_py.link_number
    dslo_cpp.link_col_obj_number = dslo_py.link_col_obj_number
    dslo_cpp.distance = dslo_py.distance
    dslo_cpp.point_link = dslo_py.point_link
    dslo_cpp.point_object = dslo_py.point_object
    dslo_cpp.jac_distance = dslo_py.jac_distance

    return dslo_cpp

def _diststructrobotobj_cpp2py(dslo_cpp, _obj, _robot):

    dsro_py = DistStructRobotObj(_obj, _robot)

    dsro_py._jac_dist_mat = np.matrix(dslo_cpp.jac_dist_mat)
    n = np.shape(dslo_cpp.dist_vect)[0]
    dsro_py._dist_vect = np.matrix(dslo_cpp.dist_vect).reshape((n,1))
    
    dsro_py._list_info = [_diststructlinkobj_cpp2py(dslo) for dslo in dslo_cpp.list_info]
    dsro_py._no_items = len(dsro_py._list_info)

    return dsro_py

def _diststructrobotobj_py2cpp(dslo_py):

    dsro_cpp = ub_cpp.CPP_DistStructRobotObj()

    dsro_cpp.is_null = dslo_py is None
    if not (dslo_py is None):
        dsro_cpp.jac_dist_mat = dslo_py.jac_dist_mat
        dsro_cpp.dist_vect = dslo_py.dist_vect
        dsro_cpp.list_info = [_diststructlinkobj_py2cpp(dslo) for dslo in dslo_py._list_info]

    return dsro_cpp



def _compute_dist(self, obj, q=None, htm=None, old_dist_struct=None, tol=0.0005, no_iter_max=20,
                  max_dist = np.inf, h=0, eps = 0, mode='auto'):
    
    n = len(self.links)

    if q is None:
        q = self.q

    if htm is None:
        htm = self.htm

    # Error handling
    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not Utils.is_a_metric_object(obj):
        raise Exception("The parameter 'obj' should be one of the following types: " + str(Utils.IS_METRIC) + ".")

    if not Utils.is_a_number(tol) or tol <= 0:
        raise Exception("The parameter 'tol' should be a positive number.")

    if not Utils.is_a_number(max_dist) or tol <= 0:
        raise Exception("The parameter 'max_dist' should be a positive number, or np.inf.")

    if not Utils.is_a_natural_number(no_iter_max) or no_iter_max <= 0:
        raise Exception("The parameter 'no_iter_max' should be a positive natural number.")

    if not Utils.is_a_number(h) or h < 0:
        raise Exception("The optional parameter 'h' must be a nonnegative number.")

    if not Utils.is_a_number(eps) or eps < 0:
        raise Exception("The optional parameter 'eps' must be a nonnegative number.")
            
            
    if not (old_dist_struct is None):
        try:
            if not (id(old_dist_struct.obj) == id(obj) and id(old_dist_struct.robot) == id(self)):
                Exception("The parameter 'old_dist_struct' is a 'DistStructRobotObj' object, but it " \
                        "must have to be relative to the SAME robot object and SAME external object, and " \
                        "this is not the case.")
        except:
            raise Exception("The parameter 'old_dist_struct' must be a 'DistStructRobotObj' object.")

    if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
        raise Exception("c++ mode is set, but .so file was not loaded!")
    # end error handling

    if mode == 'python' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
        if h >0 or eps > 0:
            raise Exception("In Python mode, smoothing parameters 'h' and 'eps' must be set to 0!")
        
        return _compute_dist_python(self, obj, Utils.cvt(q), htm, old_dist_struct, tol, no_iter_max, max_dist)
    else:
        

        old_dsro = _diststructrobotobj_py2cpp(old_dist_struct)       
        new_dsro = self.cpp_robot.compute_dist(Utils.obj_to_cpp(obj), q, htm, old_dsro, tol, no_iter_max, max_dist, h+1e-6, eps+1e-6)

        return _diststructrobotobj_cpp2py(new_dsro, obj, self)
            
        
        

def _compute_dist_python(self, obj, q=None, htm=None, old_dist_struct=None, tol=0.0005, no_iter_max=20,
                  max_dist = np.inf):
    
    n = len(self.links)

    dist_struct = DistStructRobotObj(obj, self)

    jac_dh, mth_dh = self.jac_geo(q, "dh", htm, mode = "python")

    col_object_copy = []

    # Update all collision objects of all links
    for i in range(n):
        col_object_copy.append([])
        for j in range(len(self.links[i].col_objects)):
            temp_copy = self.links[i].col_objects[j][0].copy()
            htmd = self.links[i].col_objects[j][1]
            temp_copy.set_ani_frame(mth_dh[i][:, :] * htmd)
            col_object_copy[i].append(temp_copy)

    # Compute the distance structure

    for i in range(n):
        for j in range(len(self.links[i].col_objects)):

            est_dist = 0 if math.isinf(max_dist) else Utils.compute_aabbdist(obj, col_object_copy[i][j])

            if est_dist <= max_dist:

                if old_dist_struct is None:
                    p_obj_0 = np.matrix(np.random.uniform(-100, 100, size=(3, 1)))
                else:
                    try:
                        p_obj_0 = Utils.cvt(old_dist_struct.get_item(i, j).point_object)
                    except:
                        p_obj_0 = np.matrix(np.random.uniform(-100, 100, size=(3, 1)))

                p_obj, p_obj_col, d = Utils.compute_dist(obj, col_object_copy[i][j], p_obj_0, tol, no_iter_max)

                jac_obj_col = jac_dh[i][0:3, :] - Utils.S(p_obj_col - mth_dh[i][0:3, 3]) * jac_dh[i][3:6, :]
                jac_dist = ((p_obj_col - p_obj).T * jac_obj_col) / (d+1e-6)

                dist_struct._append(i, j, d, p_obj_col, p_obj, jac_dist)



    return dist_struct



from utils import *
import numpy as np


def _jac_jac_geo(self, q=None, axis='eef', htm=None):
    if q is None:
        q = self.q
    if htm is None:
        htm = np.matrix(self.htm)

    n = len(self.links)

    # Error handling
    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    if not (axis == "eef" or axis == "dh" or axis == "com"):
        raise Exception("The parameter 'axis' should be one of the following strings:\n" \
                        "'eef': End-effector \n" \
                        "'dh': All " + str(n) + " axis of Denavit-Hartenberg\n" \
                                                "'com': All " + str(
            n) + " axis centered at the center of mass of the objects.")

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")
    # end error handling

    if (axis == 'eef') or (axis == 'dh'):
        jac_geo, htm_dh = self.jac_geo(q=q, htm=htm, axis='dh')
    else:
        jac_geo, htm_dh = self.jac_geo(q=q, htm=htm, axis='com')

    jj_geo = []

    htm0 = htm * self._htm_base_0

    for i in range(n):
        p_i = htm_dh[i][0:3, 3]
        L_i = []
        for j in range(i + 1):

            L_ij = np.matrix(np.zeros((6, n)))

            if j > 0:
                p_j_ant = htm_dh[j - 1][0:3, 3]
                z_j_ant = htm_dh[j - 1][0:3, 2]

                if self.links[j].joint_type == 0:
                    L_ij[0:3, :] = Utils.S(p_i - p_j_ant) * Utils.S(z_j_ant) @ jac_geo[j - 1][3:6, :] + Utils.S(
                        z_j_ant) * (jac_geo[i][0:3, :] - jac_geo[j - 1][0:3, :])
                    L_ij[3:6, :] = -Utils.S(z_j_ant) * jac_geo[j - 1][3:6, :]

                if self.links[j].joint_type == 1:
                    L_ij[0:3, :] = -Utils.S(z_j_ant) * jac_geo[j - 1][3:6, :]
                    # L_ij[3:6, :] = np.zeros((3, n))

            else:
                p_j_ant = htm0[0:3, 3]
                z_j_ant = htm0[0:3, 2]

                if self.links[j].joint_type == 0:
                    L_ij[0:3, :] = Utils.S(z_j_ant) * jac_geo[i][0:3, :]
                    # L_ij[3:6, :] = np.zeros((3, n))

                # if self.links[j].joint_type == 1:
                #    L_ij[0:3, :] = np.zeros((3, n))
                #    L_ij[3:6, :] = np.zeros((3, n))

            L_i.append(L_ij)

        jj_geo.append(L_i)

    if axis == 'dh' or axis == 'com':
        return jj_geo, jac_geo, htm_dh

    if axis == 'eef':
        return jj_geo[-1], jac_geo[-1], htm_dh[-1]



from .ball import *
from .box import *
from .cylinder import *
from .convexpolytope import *
from .frame import *
from .group import *
from .htmldiv import *
from .pointcloud import *
from .pointlight import *
from .rigidobject import *
from .arrow import *
